import sys
from unittest import mock

import click.exceptions
import pytest

from cumulusci.cli.robot import _is_package_installed

from .utils import run_cli_command


class MockWorkingSet(list):
    """Mocks a pkg_resources.workingset object

    Basically, a list of mock packages representing all of the packages
    installed in the current environment
    """

    def __init__(self, *package_names):
        super().__init__()
        for package_name in package_names:
            pkg = mock.Mock()
            pkg.key = package_name
            self.append(pkg)


def mock_Command(returncodes={}):
    """Create a mock for sarge.Command

    returncodes is a dictionary of commands as keys
    and returncodes as values, so we can control which
    command line tools act like they pass or fail.

    By default, every command will return zero unless
    explicitly told to exit with something else.
    """

    def the_real_mock(cmd, **kwargs):
        the_mock = mock.Mock()
        the_mock.args = cmd

        def run():
            returncode = returncodes.get(str(cmd), 0)
            the_mock.returncode = returncode

        the_mock.run.side_effect = run
        return the_mock

    return the_real_mock


@mock.patch("cumulusci.cli.robot.pkg_resources")
def test_is_package_installed(pkg_resources):
    """Verify that the helper _is_package_installed returns the correct result"""
    pkg_resources.working_set = MockWorkingSet(
        "robotframework", "robotframework-browser", "snowfakery"
    )
    result = _is_package_installed("robotframework-browser")
    assert result is True

    pkg_resources.working_set = MockWorkingSet("robotframework", "snowfakery")
    result = _is_package_installed("robotframework-browser")
    assert result is False


@mock.patch("cumulusci.cli.robot.sarge")
def test_happy_path(sarge):
    """Verify the happy path is indeed happy"""
    sarge.Command.side_effect = mock_Command()
    with mock.patch("cumulusci.cli.robot._is_package_installed", return_value=False):
        run_cli_command("robot", "install_playwright")
        actual_calls = [call.args[0] for call in sarge.Command.mock_calls]
        expected_calls = [
            "npm --version",
            [
                sys.executable,
                "-m",
                "pip",
                "install",
                "robotframework-browser",
            ],
            [sys.executable, "-m", "Browser.entry", "init"],
        ]
        msg = f"\n-- expected --\n{expected_calls}\n\n-- actual --\n{actual_calls}"

        assert actual_calls == expected_calls, msg


@mock.patch("sys.exit")
def test_bogus_subcommand(sys_exit):
    """Verify a bogus subcommand returns a UsageError"""
    with pytest.raises(click.exceptions.UsageError, match="No such command 'bogus'"):
        run_cli_command("robot", "bogus")


@mock.patch("cumulusci.cli.robot.sarge")
def test_no_npm(sarge):
    """Verify the error message we emit when npm --version throws an error"""

    sarge.Command.side_effect = mock_Command(returncodes={"npm --version": 1})
    with pytest.raises(
        click.exceptions.ClickException, match="Unable to find a usable npm.*"
    ):
        run_cli_command("robot", "install_playwright", "--dry_run")
    sarge.Command.assert_called_once()
    assert sarge.Command.mock_calls[0].args[0] == "npm --version"


@mock.patch("cumulusci.cli.robot.sarge")
def test_playwright_dry_run(sarge):
    """Verify the output of a dry run"""
    sarge.Command.side_effect = mock_Command(returncodes={"npm --version": 0})
    with mock.patch("cumulusci.cli.robot._is_package_installed", return_value=False):
        result = run_cli_command("robot", "install_playwright", "--dry_run")
        expected_output = "\n".join(
            [
                "installing robotframework-browser ...",
                f"would run {sys.executable} -m pip install robotframework-browser",
                f"would run {sys.executable} -m Browser.entry init\n",
            ]
        )
        msg = f"\n-- expected --\n{expected_output}\n\n-- actual --\n{result.output}"
        assert result.output == expected_output, msg


@mock.patch("cumulusci.cli.robot.sarge")
def test_playwright_failure_to_initialize_browser_library(sarge):
    """Verify we raise an exception if we can't initialize the browser library"""
    cmd = str([sys.executable, "-m", "Browser.entry", "init"])
    sarge.Command.side_effect = mock_Command(returncodes={cmd: 1})
    with mock.patch("cumulusci.cli.robot._is_package_installed", return_value=False):
        with pytest.raises(
            click.exceptions.ClickException,
            match="unable to initialize browser library",
        ):
            run_cli_command("robot", "install_playwright")


@mock.patch("cumulusci.cli.robot.sarge")
def test_playwright_failure_to_install_browser_library(sarge):
    """Verify we raise an exception if we can't install playwright"""
    cmd = str([sys.executable, "-m", "pip", "install", "robotframework-browser"])
    sarge.Command.side_effect = mock_Command(returncodes={cmd: 1})
    with mock.patch("cumulusci.cli.robot._is_package_installed", return_value=False):
        with pytest.raises(
            click.exceptions.ClickException,
            match="robotframework-browser was not installed",
        ):
            run_cli_command("robot", "install_playwright")


@mock.patch("cumulusci.cli.robot.sarge")
def test_playwright_already_installed(sarge):
    """Verify we exit early if the package is already installed"""
    sarge.Command.side_effect = mock_Command()
    with mock.patch("cumulusci.cli.robot._is_package_installed", return_value=True):
        result = run_cli_command("robot", "install_playwright", "--dry_run")
        assert (
            result.output == "Playwright support seems to already have been installed\n"
        )


@mock.patch("cumulusci.cli.robot.sarge")
def test_uninstall_playwright(sarge):
    """Verify calling 'robot uninstall_playwright' calls the right utilities"""
    sarge.Command.side_effect = mock_Command()
    with mock.patch("cumulusci.cli.robot._is_package_installed", return_value=False):
        run_cli_command("robot", "uninstall_playwright")
        sarge.Command.assert_has_calls(
            [
                mock.call([sys.executable, "-m", "Browser.entry", "clean-node"]),
                mock.call(
                    [
                        sys.executable,
                        "-m",
                        "pip",
                        "uninstall",
                        "robotframework-browser",
                        "--yes",
                    ],
                ),
            ]
        )



from contextlib import contextmanager

from cumulusci.robotframework.base_library import BaseLibrary


class BasePage(BaseLibrary):
    _object_name = None

    def __init__(self, object_name=None):
        super().__init__()
        if object_name:
            self._object_name = object_name

    @contextmanager
    def _no_implicit_wait(self):
        """Context manager for running selenium commands without an implicit wait"""
        current_wait = self.selenium.set_selenium_implicit_wait(0)
        try:
            yield
        finally:
            self.selenium.set_selenium_implicit_wait(current_wait)

    def _wait_to_appear(self, *args, timeout="15s"):
        """This is the concrete implementation for the 'Wait for page object' keyword

        It can be overridden by a subclass to do whatever is necessary to
        wait for itself to appear.
        """

        try:
            self.selenium.wait_for_condition(
                "return (document.readyState == 'complete')"
            )
            self.salesforce.wait_for_aura()
            self.builtin.wait_until_keyword_succeeds(
                timeout,
                "2s",
                "Current page should be",
                self._page_type,
                self._object_name,
            )
        except Exception as e:
            self.builtin.log(e, "DEBUG")
            raise Exception(
                f"Page object {self._page_type} {self._object_name} did not appear before timeout ({timeout}) expired."
            )

    def _remove_from_library_search_order(self):
        """Remove the current page object from robot's library search order

        Note: robot doesn't provide a way to completely unload a
        library during execution. However, this at least makes sure that
        all other libraries will have priority over this one.
        """
        order = list(self.builtin.set_library_search_order())
        if self._libname in order:
            order.remove(self._libname)
            self.builtin.set_library_search_order(*order)
            self.builtin.log("new search order: {}".format(order), "DEBUG")

    @property
    def object_name(self):
        object_name = self._object_name

        # the length check is to skip objects from a different namespace
        # like foobar__otherpackageobject__c
        if object_name is not None:
            parts = object_name.split("__")
            if len(parts) == 2 and parts[-1] == "c":
                # get_namespace_prefix already takes care of returning an actual
                # prefix or an empty string depending on whether the package is managed
                object_name = "{}{}".format(
                    self.cumulusci.get_namespace_prefix(), object_name
                )
        return object_name

    @property
    def selenium(self):
        """Returns the instance of the imported SeleniumLibrary library"""
        return self.builtin.get_library_instance("SeleniumLibrary")

    def _get_object(self, **kwargs):
        """Get the object associated with the given keyword arguments.

        This performs a salesforce query. It will raise an exception unless
        exactly one result is returned from the query.
        """
        results = self.salesforce_api.salesforce_query(self.object_name, **kwargs)
        if len(results) == 0:
            human_friendly_args = ", ".join(
                ["{}={}".format(key, kwargs[key]) for key in kwargs]
            )
            raise Exception(
                "no {} matches {}".format(self.object_name, human_friendly_args)
            )
        elif len(results) > 1:
            raise Exception("Query returned {} objects".format(len(results)))
        else:
            return results[0]

    def log_current_page_object(self):
        """Logs the name of the current page object

        The current page object is also returned as an object
        """
        polib = self.builtin.get_library_instance(
            "cumulusci.robotframework.PageObjects"
        )
        if polib.current_page_object is None:
            # this should not be possible, since this keyword is only available
            # if you've loaded a page object. Still, better safe then sorry.
            self.builtin.log("no page object has been loaded")
            return None
        else:
            pobj = polib.current_page_object

            self.builtin.log(
                "current page object: {}".format(polib.current_page_object)
            )
            return pobj



from cumulusci.salesforce_api.metadata import ApiRetrievePackaged
from cumulusci.tasks.salesforce import BaseRetrieveMetadata
from cumulusci.utils import zip_subfolder

retrieve_packaged_options = BaseRetrieveMetadata.task_options.copy()
retrieve_packaged_options.update(
    {
        "package": {
            "description": "The package name to retrieve.  Defaults to project__package__name",
            "required": True,
        },
        "api_version": {
            "description": (
                "Override the default api version for the retrieve."
                + " Defaults to project__package__api_version"
            )
        },
    }
)


class RetrievePackaged(BaseRetrieveMetadata):
    api_class = ApiRetrievePackaged

    task_options = retrieve_packaged_options

    def _init_options(self, kwargs):
        super(RetrievePackaged, self)._init_options(kwargs)
        if "package" not in self.options:
            self.options["package"] = self.project_config.project__package__name

    def _get_api(self):
        return self.api_class(
            self, self.options["package"], self.options.get("api_version")
        )

    def _extract_zip(self, src_zip):
        src_zip = zip_subfolder(src_zip, self.options.get("package"))
        super(RetrievePackaged, self)._extract_zip(src_zip)



import datetime
import json
import os
import re
import sys
import tempfile
from pathlib import Path
from unittest import mock

import pytest

from cumulusci.core import utils
from cumulusci.core.config import (
    BaseConfig,
    OrgConfig,
    ScratchOrgConfig,
    ServiceConfig,
    UniversalConfig,
)
from cumulusci.core.config.marketing_cloud_service_config import (
    MarketingCloudServiceConfig,
)
from cumulusci.core.config.sfdx_org_config import SfdxOrgConfig
from cumulusci.core.exceptions import (
    ConfigError,
    CumulusCIException,
    CumulusCIUsageError,
    KeychainKeyNotFound,
    OrgCannotBeLoaded,
    OrgNotFound,
    ServiceCannotBeLoaded,
    ServiceNotConfigured,
    ServiceNotValid,
)
from cumulusci.core.keychain import EncryptedFileProjectKeychain
from cumulusci.core.keychain.base_project_keychain import (
    DEFAULT_CONNECTED_APP,
    DEFAULT_CONNECTED_APP_NAME,
)
from cumulusci.core.keychain.encrypted_file_project_keychain import (
    SERVICE_ORG_FILE_MODE,
    GlobalOrg,
)
from cumulusci.core.keychain.serialization import (
    _simplify_config,
    load_config_from_json_or_pickle,
)
from cumulusci.core.tests.utils import EnvironmentVarGuard
from cumulusci.utils import temporary_dir


@pytest.fixture(params=[True, False])
def withdifferentformats(request):
    SHOULD_SAVE_AS_JSON = request.param
    with mock.patch(
        "cumulusci.core.keychain.serialization.SHOULD_SAVE_AS_JSON", request.param
    ):
        yield "json" if SHOULD_SAVE_AS_JSON else "pickle"


@pytest.fixture()
def keychain(project_config, key) -> EncryptedFileProjectKeychain:
    keychain = EncryptedFileProjectKeychain(project_config, key)
    assert keychain.project_config == project_config
    assert keychain.key == key
    return keychain


class TestEncryptedFileProjectKeychain:

    project_name = "TestProject"

    def _write_file(self, filepath, contents):
        with open(filepath, "w") as f:
            f.write(contents)

    def _default_org_path(self, keychain: EncryptedFileProjectKeychain):
        return Path(keychain.global_config_dir, "TestProject/DEFAULT_ORG.txt")

    def test_cache_dir(self, keychain):
        assert keychain.cache_dir.name == ".cci"

    def test_raise_service_not_configured(self, keychain):
        with pytest.raises(ServiceNotConfigured):
            keychain._raise_service_not_configured("test-name")

    #######################################
    #               Orgs                  #
    #######################################

    def test_set_and_get_org__global(
        self, keychain, org_config, key, withdifferentformats
    ):
        org_config.global_org = True
        keychain.set_org(org_config, True)
        assert list(keychain.orgs.keys()) == ["test"]
        assert _simplify_config(keychain.get_org("test").config) == org_config.config

    def test_get_org__with_config_properly_overridden(
        self, keychain, scratch_org_config
    ):
        days = 16
        config_file = "./foo/bar/baz"
        scratch_org_config.global_org = True
        # the orgs encrypted file has the default value for days and config_file
        keychain.set_org(scratch_org_config, True)
        # but this particular scratch org has days and config_file specified via cumulusci.yml
        keychain.project_config.config = {
            "orgs": {"scratch": {"test": {"days": days, "config_file": config_file}}}
        }
        org = keychain.get_org("test")

        # ensure what is configured in cumulusci.yml is what is loaded into the config
        assert org.config["days"] == days
        assert org.config["config_file"] == config_file

    def test_get_org__not_found(self, keychain):
        org_name = "mythical"
        error_message = f"Org with name '{org_name}' does not exist."
        with pytest.raises(OrgNotFound, match=error_message):
            keychain.get_org(org_name)

    def test_set_org__no_key_should_save_to_unencrypted_file(
        self, keychain, org_config
    ):
        keychain.key = None
        keychain.set_org(org_config)

        filepath = Path(keychain.project_local_dir, "test.org")
        with open(filepath, "rb") as f:
            assert load_config_from_json_or_pickle(f.read()) == {
                **org_config.config,
                # still the default for now
                "serialization_format": "pickle",
            }

    def test_set_org__should_not_save_when_environment_project_keychain_set(
        self, keychain, org_config, withdifferentformats
    ):
        with temporary_dir() as temp:
            env = EnvironmentVarGuard()
            with EnvironmentVarGuard() as env:
                env.set("CUMULUSCI_KEYCHAIN_CLASS", "EnvironmentProjectKeychain")
                with mock.patch.object(
                    EncryptedFileProjectKeychain, "project_local_dir", temp
                ):
                    keychain.set_org(org_config, global_org=False)

            actual_org = keychain.get_org("test")
            assert actual_org.config == {
                **org_config.config,
                "serialization_format": withdifferentformats,
            }
            assert not Path(temp, "test.org").is_file()

    @mock.patch("cumulusci.core.keychain.encrypted_file_project_keychain.open")
    def test_save_org_when_no_project_local_dir_present(
        self, mock_open, keychain, org_config
    ):
        with mock.patch.object(EncryptedFileProjectKeychain, "project_local_dir", None):
            keychain._save_org("alias", org_config, global_org=False)
        assert mock_open.call_count == 0

    def test_load_files__org_empty(self, keychain):
        self._write_file(
            Path(keychain.global_config_dir, "test.org"),
            keychain._get_config_bytes(BaseConfig({"foo": "bar"})).decode("utf-8"),
        )

        del keychain.config["orgs"]
        keychain._load_orgs()
        assert "foo" in keychain.get_org("test").config
        assert keychain.get_org("test").keychain == keychain

    def test_remove_org(self, keychain, org_config):
        keychain.set_org(org_config)
        keychain.remove_org("test")
        assert "test" not in keychain.orgs

    def test_remove_org__not_found(self, keychain):
        keychain.orgs["test"] = mock.Mock()
        with pytest.raises(OrgNotFound):
            keychain.remove_org("test")

    def test_remove_org__global_not_found(self, keychain):
        keychain.orgs["test"] = mock.Mock()
        with pytest.raises(OrgNotFound):
            keychain.remove_org("test", True)

    def test_set_and_get_org_local_should_not_shadow_global(
        self, keychain, org_config, project_config, key, withdifferentformats
    ):
        global_org = True
        org_config.global_org = global_org
        keychain.set_org(org_config, global_org)
        assert ["test"] == list(keychain.orgs.keys())
        assert isinstance(keychain.orgs["test"], GlobalOrg), keychain.orgs["test"]
        test_config = keychain.get_org("test").config
        assert {
            **org_config.config,
            "serialization_format": withdifferentformats,
        } == test_config

        org_filepath = Path(keychain.global_config_dir, "test.org")
        assert org_filepath.exists()

        # os.stat returns something different on windows
        # so only check this on osx/linux
        if not sys.platform.startswith("win"):
            # ensure expected file permissions
            stat_result = os.stat(org_filepath)
            actual_mode = oct(stat_result.st_mode & 0o777)
            assert actual_mode == oct(SERVICE_ORG_FILE_MODE)

        # check that it saves to the right place
        with mock.patch(
            "cumulusci.core.keychain.encrypted_file_project_keychain.os.open"
        ) as o:
            org_config.save()
            opened_file = o.mock_calls[0][1][0]
            assert opened_file.parent.name == ".cumulusci"
            assert opened_file.name == "test.org"

        # check that it can be loaded in a fresh keychain
        new_keychain = EncryptedFileProjectKeychain(project_config, key)
        org_config = new_keychain.get_org("test")
        assert org_config.global_org

    def test_get_default_org__with_files(self, keychain, org_config):
        org_config = OrgConfig(org_config.config.copy(), "test", keychain=keychain)
        org_config.save()
        with open(self._default_org_path(keychain), "w") as f:
            f.write("test")
        try:
            assert (
                _simplify_config(keychain.get_default_org()[1].config)
                == org_config.config
            )
        finally:
            self._default_org_path(keychain).unlink()

    def test_get_default_org__with_files__missing_org(self, keychain):
        with open(self._default_org_path(keychain), "w") as f:
            f.write("should_not_exist")
        assert self._default_org_path(keychain).exists()
        assert keychain.get_default_org() == (None, None)
        assert not self._default_org_path(keychain).exists()

    @mock.patch("sarge.Command")
    def test_set_default_org__with_files(self, Command, keychain, org_config):
        org_config = OrgConfig(org_config.config.copy(), "test")
        keychain.set_org(org_config)
        keychain.set_default_org("test")
        with open(self._default_org_path(keychain)) as f:
            assert f.read() == "test"
        self._default_org_path(keychain).unlink()

    @mock.patch("sarge.Command")
    def test_unset_default_org__with_files(self, Command, keychain, org_config):
        org_config = org_config.config.copy()
        org_config = OrgConfig(org_config, "test")
        keychain.set_org(org_config)
        keychain.set_default_org("test")
        keychain.unset_default_org()
        assert keychain.get_default_org()[1] is None
        assert not self._default_org_path(keychain).exists()

    # old way of finding defaults used contents of the files themselves
    # we should preserve backwards compatibiliity for a few months
    def test_get_default_org__file_missing_fallback(self, keychain, org_config):
        org_config = OrgConfig(org_config.config.copy(), "test", keychain=keychain)
        org_config.config["default"] = True
        org_config.save()
        assert (
            _simplify_config(keychain.get_default_org()[1].config) == org_config.config
        )

    def test_get_default_org__outside_project(self, keychain):
        assert keychain.get_default_org() == (None, None)

    def test_load_orgs_from_environment(self, keychain, org_config):
        scratch_config = org_config.config.copy()
        scratch_config["scratch"] = True
        env = EnvironmentVarGuard()
        with EnvironmentVarGuard() as env:
            env.set(
                f"{keychain.env_org_var_prefix}dev",
                json.dumps(scratch_config),
            )
            env.set(
                f"{keychain.env_org_var_prefix}devhub",
                json.dumps(org_config.config),
            )
            keychain._load_orgs_from_environment()

        actual_config = keychain.get_org("dev")
        assert _simplify_config(actual_config.config) == scratch_config
        actual_config = keychain.get_org("devhub")
        assert _simplify_config(actual_config.config) == org_config.config

    def test_load_orgs_from_environment__empty_throws_error(self, keychain, org_config):
        env = EnvironmentVarGuard()
        with EnvironmentVarGuard() as env:
            env.set(
                f"{keychain.env_org_var_prefix}dev",
                "",
            )
            with pytest.raises(OrgCannotBeLoaded):
                keychain._load_orgs_from_environment()

    def test_load_orgs_from_environment__invalid_json_throws_error(
        self, keychain, org_config
    ):
        env = EnvironmentVarGuard()
        with EnvironmentVarGuard() as env:
            env.set(
                f"{keychain.env_org_var_prefix}dev",
                "['foo',]",
            )
            with pytest.raises(OrgCannotBeLoaded):
                keychain._load_orgs_from_environment()

    #######################################
    #              Services               #
    #######################################

    def test_load_service_files(self, keychain):
        github_service_path = Path(f"{keychain.global_config_dir}/services/github")
        self._write_file(
            Path(github_service_path / "alias.service"),
            keychain._get_config_bytes(BaseConfig({"name": "foo"})).decode("utf-8"),
        )

        # keychain.config["services"] = {}

        with mock.patch.object(
            EncryptedFileProjectKeychain,
            "global_config_dir",
            keychain.global_config_dir,
        ):
            keychain._load_service_files()
        github_service = keychain.get_service("github", "alias")
        assert "foo" in github_service.config["name"]

    def test_load_services__from_env(self, keychain):
        service_config_one = ServiceConfig(
            {"name": "foo1", "password": "1234", "token": "1234"}
        )
        service_config_two = ServiceConfig(
            {"name": "foo2", "password": "5678", "token": "5678"}
        )
        with EnvironmentVarGuard() as env:
            env.set(
                f"{keychain.env_service_var_prefix}github",
                json.dumps(service_config_one.config),
            )
            env.set(
                f"{keychain.env_service_var_prefix}github__OTHER",
                json.dumps(service_config_two.config),
            )
            with pytest.raises(ServiceNotConfigured):
                keychain.get_service("github")

            keychain._load_services_from_environment()

        gh_service = keychain.get_service("github")
        # this also confirms the default service is set appropriately
        assert _simplify_config(gh_service.config) == service_config_one.config
        gh_service = keychain.get_service("github", "env-OTHER")
        assert _simplify_config(gh_service.config) == service_config_two.config

    def test_load_services_from_env__same_name_throws_error(self, keychain):
        keychain.logger = mock.Mock()
        service_prefix = EncryptedFileProjectKeychain.env_service_var_prefix
        service_config = ServiceConfig(
            {"name": "foo", "password": "1234", "token": "1234"}
        )
        with EnvironmentVarGuard() as env:
            env.set(f"{service_prefix}github", json.dumps(service_config.config))
            env.set(f"{service_prefix}github", json.dumps(service_config.config))
            keychain._load_services_from_environment()

        assert 1 == 1

    def test_load_services_from_env__empty_throws_error(self, keychain):
        service_prefix = EncryptedFileProjectKeychain.env_service_var_prefix
        env = EnvironmentVarGuard()
        with EnvironmentVarGuard() as env:
            env.set(
                f"{service_prefix}github",
                "",
            )
            with pytest.raises(ServiceCannotBeLoaded):
                keychain._load_services_from_environment()

    def test_load_services_from_env__invalid_json_throws_error(self, keychain):
        service_prefix = EncryptedFileProjectKeychain.env_service_var_prefix
        env = EnvironmentVarGuard()
        with EnvironmentVarGuard() as env:
            env.set(
                f"{service_prefix}github",
                "['foo',]",
            )
            with pytest.raises(ServiceCannotBeLoaded):
                keychain._load_services_from_environment()

    def test_get_service__built_in_connected_app(self, keychain):
        built_in_connected_app = keychain.get_service("connected_app")
        assert built_in_connected_app is DEFAULT_CONNECTED_APP

    def test_get_service__with_class_path(self, keychain, service_config):
        encrypted = keychain._get_config_bytes(service_config)
        keychain.config["services"]["marketing_cloud"] = {"foo": encrypted}
        mc_service = keychain._get_service("marketing_cloud", "foo")

        assert isinstance(mc_service, MarketingCloudServiceConfig)

    @mock.patch("cumulusci.core.keychain.encrypted_file_project_keychain.import_class")
    def test_get_service__bad_class_path(self, import_class, keychain, service_config):
        import_class.side_effect = AttributeError
        encrypted = keychain._get_config_bytes(service_config)
        keychain.config["services"]["marketing_cloud"] = {"foo": encrypted}

        error_message = "Unrecognized class_path for service: cumulusci.core.config.marketing_cloud_service_config.MarketingCloudServiceConfig"
        with pytest.raises(CumulusCIException, match=error_message):
            keychain._get_service("marketing_cloud", "foo")

    def test_get_service__does_not_exist(self, keychain, service_config):
        keychain.set_service("github", "alias", service_config)
        error_message = "No service of type github exists with the name: does-not-exist"
        with pytest.raises(ServiceNotConfigured, match=error_message):
            keychain.get_service("github", "does-not-exist")

    @pytest.mark.skipif(
        sys.platform.startswith("win"),
        reason="Windows has a different value returned by os.stat",
    )
    def test_set_service_github(self, keychain, service_config):
        keychain.set_service("github", "alias", service_config)
        default_github_service = keychain.get_service("github")

        service_filepath = Path(
            keychain.global_config_dir, "services/github/alias.service"
        )
        assert service_filepath.is_file()

        if not sys.platform.startswith("win"):
            # ensure expected file permissions
            stat_result = os.stat(service_filepath)
            actual_mode = oct(stat_result.st_mode & 0o777)
            assert actual_mode == oct(SERVICE_ORG_FILE_MODE)

        assert default_github_service.config == {
            **service_config.config,
            "token": "test123",
            "serialization_format": "pickle",
        }

    @pytest.mark.skipif(
        not sys.platform.startswith("win"),
        reason="Windows has a different value returned by os.stat",
    )
    def test_set_service_github__windows_has_correct_file_perms(
        self, keychain, service_config
    ):
        keychain.set_service("github", "alias", service_config)

        service_filepath = Path(
            keychain.global_config_dir, "services/github/alias.service"
        )
        assert service_filepath.is_file()

        # ensure expected file permissions
        stat_result = os.stat(service_filepath)
        actual_mode = oct(stat_result.st_mode & 0o777)
        assert actual_mode == "0o666"

    def test_set_service__cannot_overwrite_default_connected_app(self, keychain):
        connected_app_config = ServiceConfig({"test": "foo"})
        error_message = re.escape(
            f"You cannot use the name {DEFAULT_CONNECTED_APP_NAME} for a connected app service. Please select a different name."
        )
        with pytest.raises(ServiceNotValid, match=error_message):
            keychain.set_service(
                "connected_app", DEFAULT_CONNECTED_APP_NAME, connected_app_config
            )

    def test_set_service__first_should_be_default(self, keychain):
        keychain.set_service("github", "foo_github", ServiceConfig({"name": "foo"}))
        keychain.set_service("github", "bar_github", ServiceConfig({"name": "bar"}))

        github_service = keychain.get_service("github")
        assert _simplify_config(github_service.config) == {"name": "foo"}

    def test_set_default_service(self, keychain, withdifferentformats):
        keychain.set_service("github", "foo_github", ServiceConfig({"name": "foo"}))
        keychain.set_service("github", "bar_github", ServiceConfig({"name": "bar"}))

        github_service = keychain.get_service("github")
        assert _simplify_config(github_service.config) == {"name": "foo"}
        # now set default to bar
        keychain.set_default_service("github", "bar_github")
        github_service = keychain.get_service("github")
        assert _simplify_config(github_service.config) == {"name": "bar"}

    def test_set_default_service__service_alredy_default(
        self, keychain, withdifferentformats
    ):
        keychain.set_service("github", "foo_github", ServiceConfig({"name": "foo"}))
        github_service = keychain.get_service("github")
        assert github_service.config == {
            "name": "foo",
            "serialization_format": withdifferentformats,
        }

        keychain.set_default_service("github", "foo_github")

        github_service = keychain.get_service("github")
        assert github_service.config == {
            "name": "foo",
            "serialization_format": withdifferentformats,
        }

    def test_set_default_service__no_such_service(self, keychain):
        with pytest.raises(ServiceNotConfigured):
            keychain.set_default_service("fooey", "alias")

    def test_set_default_service__no_such_alias(self, keychain):
        keychain.set_service("github", "foo_github", ServiceConfig({"name": "foo"}))
        with pytest.raises(ServiceNotConfigured):
            keychain.set_default_service("github", "wrong_alias")

    def test_save_default_service__global_default_service(self, keychain):
        with open(Path(keychain.global_config_dir, "DEFAULT_SERVICES.json"), "w") as f:
            f.write(
                json.dumps({"devhub": "current_default", "github": "current_default"})
            )

        keychain._save_default_service("github", "new_default", project=False)

        with open(Path(keychain.global_config_dir, "DEFAULT_SERVICES.json"), "r") as f:
            default_services = json.loads(f.read())

        assert default_services["devhub"] == "current_default"
        assert default_services["github"] == "new_default"

        with open(Path(keychain.project_local_dir, "DEFAULT_SERVICES.json"), "r") as f:
            project_defaults = json.loads(f.read())

        assert project_defaults == {}

    def test_save_default_service__project_default_service(self, keychain):
        with open(Path(keychain.project_local_dir, "DEFAULT_SERVICES.json"), "w") as f:
            f.write(
                json.dumps({"devhub": "current_default", "github": "current_default"})
            )

        keychain._save_default_service("github", "new_default", project=True)

        with open(Path(keychain.project_local_dir, "DEFAULT_SERVICES.json"), "r") as f:
            default_services = json.loads(f.read())

        assert default_services["devhub"] == "current_default"
        assert default_services["github"] == "new_default"

        with open(Path(keychain.global_config_dir, "DEFAULT_SERVICES.json"), "r") as f:
            global_defaults = json.loads(f.read())

        assert global_defaults == {}

    def test_iter_local_project_dirs(self, keychain):
        cci_home_dir = Path(keychain.global_config_dir)
        (cci_home_dir / "logs").mkdir()
        (cci_home_dir / "chewy").mkdir()
        (cci_home_dir / "yoshi").mkdir()
        # services/ and TestProject/ are also presenct in ~/.cumulusci

        local_project_dirs = list(keychain._iter_local_project_dirs())
        assert cci_home_dir / "TestProject" in local_project_dirs
        assert cci_home_dir / "chewy" in local_project_dirs
        assert cci_home_dir / "yoshi" in local_project_dirs

    def test_create_default_service_file__file_already_exists(self, key):
        home_dir = tempfile.mkdtemp()
        cci_home_dir = Path(f"{home_dir}/.cumulusci")
        cci_home_dir.mkdir(parents=True)
        local_project_dir = cci_home_dir / "test-project"
        local_project_dir.mkdir()
        self._write_file(cci_home_dir / "DEFAULT_SERVICES.json", json.dumps({}))
        with mock.patch.object(
            EncryptedFileProjectKeychain, "global_config_dir", cci_home_dir
        ):
            keychain = EncryptedFileProjectKeychain(UniversalConfig(), key)
            keychain._create_default_service_files()

        assert not (local_project_dir / "DEFAULT_SERVICE.json").is_file()

    def test_create_default_services_files__without_project_service(self, key):
        home_dir = tempfile.mkdtemp()
        cci_home_dir = Path(f"{home_dir}/.cumulusci")
        cci_home_dir.mkdir(parents=True)

        self._write_file(cci_home_dir / "devhub.service", "<encrypted devhub config>")
        self._write_file(cci_home_dir / "github.service", "<encrypted github config>")

        # local project dir without a .service file
        (cci_home_dir / "test-project").mkdir()
        # we should ignore everything in the log dir
        log_dir = cci_home_dir / "logs"
        log_dir.mkdir()
        self._write_file(log_dir / "connected_app.service", "<encrypted config>")

        with mock.patch.object(
            EncryptedFileProjectKeychain, "global_config_dir", cci_home_dir
        ):
            # _create_default_services_files() invoked via __init__
            EncryptedFileProjectKeychain(UniversalConfig(), key)

        default_services_file = cci_home_dir / "DEFAULT_SERVICES.json"
        with open(default_services_file, "r") as f:
            default_services = json.loads(f.read())

        assert len(default_services.keys()) == 2  # we shouldn't get connected_app
        assert default_services["devhub"] == "global"
        assert default_services["github"] == "global"

        project_default_services_file = Path(
            f"{cci_home_dir}/test-project/DEFAULT_SERVICES.json"
        )
        with open(project_default_services_file, "r") as f:
            assert json.loads(f.read()) == {}

    def test_create_default_services_files__with_project_service(self, key):
        home_dir = tempfile.mkdtemp()
        cci_home_dir = Path(f"{home_dir}/.cumulusci")
        cci_home_dir.mkdir(parents=True)

        self._write_file(cci_home_dir / "devhub.service", "<encrypted devhub config>")
        self._write_file(cci_home_dir / "github.service", "<encrypted github config>")

        project_path = cci_home_dir / "test-project"
        project_path.mkdir(parents=True)
        self._write_file(project_path / "github.service", "project level github config")

        with mock.patch.object(
            EncryptedFileProjectKeychain, "global_config_dir", cci_home_dir
        ):
            with mock.patch.object(
                EncryptedFileProjectKeychain, "project_local_dir", project_path
            ):
                # _create_default_services_files invoked via __init__
                EncryptedFileProjectKeychain(UniversalConfig(), key)

        global_default_services_file = Path(f"{cci_home_dir}/DEFAULT_SERVICES.json")
        assert global_default_services_file.is_file()
        with open(global_default_services_file, "r") as f:
            global_defaults = json.loads(f.read())

        assert len(global_defaults.keys()) == 2
        assert global_defaults["github"] == "global"
        assert global_defaults["devhub"] == "global"

        project_default_services_file = project_path / "DEFAULT_SERVICES.json"
        assert project_default_services_file.is_file()
        with open(project_default_services_file, "r") as f:
            default_services = json.loads(f.read())

        assert len(default_services.keys()) == 1
        assert default_services["github"] == "test-project"

    def test_create_services_dir_structure(self, key):
        service_types = list(UniversalConfig().config["services"].keys())
        num_services = len(service_types)

        # _create_services_dir_structure() is invoked via constructor
        keychain = EncryptedFileProjectKeychain(UniversalConfig(), key)

        services_path = Path(f"{keychain.global_config_dir}/services")
        for path in Path.iterdir(services_path):
            if path.name in service_types:
                assert Path.is_dir(path)
                service_types.remove(path.name)

        assert len(service_types) == 0

        # explicitly invoke a second time to test idempotency
        keychain._create_services_dir_structure()
        # make sure no new dirs appeared
        assert num_services == len(list(Path.iterdir(services_path)))

    def test_migrate_services(self, key, project_config):
        home_dir = tempfile.mkdtemp()
        cci_home_dir = Path(f"{home_dir}/.cumulusci")
        cci_home_dir.mkdir(parents=True)

        self._write_file(cci_home_dir / "github.service", "github config")
        self._write_file(cci_home_dir / "foo.service", "foo config")

        local_proj_dir = cci_home_dir / "test-project"
        local_proj_dir.mkdir()
        self._write_file(local_proj_dir / "github.service", "github2 config")

        with mock.patch.object(
            EncryptedFileProjectKeychain, "global_config_dir", cci_home_dir
        ):
            # _migrade_services() invoked via __init__
            EncryptedFileProjectKeychain(project_config, key)

        assert not Path.is_file(cci_home_dir / "github.service")

        new_github_service_file = cci_home_dir / "services/github/global.service"
        assert new_github_service_file.is_file()

        # ensure expected file contents
        with open(cci_home_dir / "services/github/global.service") as f:
            assert f.read() == "github config"

        assert not Path.is_file(cci_home_dir / "test-project/devhub.service")
        assert (cci_home_dir / "services/github/TestProject.service").is_file()
        with open(cci_home_dir / "services/github/TestProject.service") as f:
            assert f.read() == "github2 config"

        # unrecognized services should be left alone
        assert (cci_home_dir / "foo.service").is_file()

    def test_migrate_services__warn_duplicate_default_service(
        self, project_config, key
    ):
        home_dir = tempfile.mkdtemp()
        cci_home_dir = Path(f"{home_dir}/.cumulusci")
        cci_home_dir.mkdir(parents=True)

        # make unnamed devhub service
        legacy_devhub_service = Path(f"{cci_home_dir}/devhub.service")
        self._write_file(legacy_devhub_service, "legacy config")
        # make existing default aliased devhub service
        named_devhub_service = Path(f"{cci_home_dir}/services/devhub/")
        named_devhub_service.mkdir(parents=True)
        self._write_file(f"{named_devhub_service}/global.service", "migrated config")

        with mock.patch.object(
            EncryptedFileProjectKeychain, "global_config_dir", cci_home_dir
        ):
            # _migrate_services() invoked via __init__
            keychain = EncryptedFileProjectKeychain(project_config, key)
            keychain._migrate_services_from_dir(cci_home_dir)

        # ensure this file wasn't removed
        assert legacy_devhub_service.is_file()
        # ensure contents of migrated are unchanged
        with open(named_devhub_service / "global.service", "r") as f:
            assert f.read() == "migrated config"

    def test_rename_service(self, keychain, service_config, withdifferentformats):
        home_dir = tempfile.mkdtemp()

        cci_home_dir = Path(f"{home_dir}/.cumulusci")
        cci_home_dir.mkdir()
        with open(cci_home_dir / "DEFAULT_SERVICES.json", "w") as f:
            f.write(json.dumps({"github": "old_alias"}))

        local_project_dir = cci_home_dir / "test-project"
        local_project_dir.mkdir()
        with open(local_project_dir / "DEFAULT_SERVICES.json", "w") as f:
            f.write(json.dumps({"github": "old_alias"}))

        github_services_dir = Path(f"{home_dir}/.cumulusci/services/github")
        github_services_dir.mkdir(parents=True)
        self._write_file(github_services_dir / "old_alias.service", "github config")

        encrypted = keychain._get_config_bytes(service_config)
        keychain.services = {"github": {"old_alias": encrypted}}
        with mock.patch.object(
            EncryptedFileProjectKeychain, "global_config_dir", cci_home_dir
        ):
            keychain.rename_service("github", "old_alias", "new_alias")

        # Getting old alias should fail
        with pytest.raises(ServiceNotConfigured):
            keychain.get_service("github", "old_alias")

        # Validate new alias has same contents as original
        assert keychain.get_service("github", "new_alias").config == {
            **service_config.config,
            "token": "test123",
            "serialization_format": withdifferentformats,
        }
        # Old service file should be gone
        assert not (github_services_dir / "old_alias.service").is_file()
        # New service file should be present
        assert (github_services_dir / "new_alias.service").is_file()

        # DEFAULT_SERVICES.json files should have updated aliases
        with open(local_project_dir / "DEFAULT_SERVICES.json", "r") as f:
            assert json.loads(f.read()) == {"github": "new_alias"}

        with open(cci_home_dir / "DEFAULT_SERVICES.json", "r") as f:
            assert json.loads(f.read()) == {"github": "new_alias"}

    def test_rename_service__new_alias_already_exists(self, keychain):
        keychain.services = {
            "github": {"old-alias": "old config", "new-alias": "new config"}
        }
        with pytest.raises(
            CumulusCIUsageError,
            match="A service of type github already exists with name: new-alias",
        ):
            keychain.rename_service("github", "old-alias", "new-alias")

    def test_rename_service__invalid_service_type(self, keychain, service_config):
        keychain.services = {"github": {"alias": "config"}}
        with pytest.raises(ServiceNotConfigured):
            keychain.rename_service("does-not-exist", "old-alias", "new-alias")

    def test_rename_service__invalid_service_alias(self, keychain, service_config):
        keychain.services = {"github": {"current_alias": "some config"}}
        with pytest.raises(ServiceNotConfigured):
            keychain.rename_service("github", "does-not-exist", "new_alias")

    def test_rename_alias_in_default_service_file__no_default_present(self, keychain):
        filepath = Path(f"{keychain.global_config_dir}/DEFAULT_SERVICES.json")
        with open(filepath, "w") as f:
            f.write(json.dumps({"saucelabs": "default_alias"}))
        keychain._rename_alias_in_default_service_file(
            filepath, "github", "current_alias", "new_alias"
        )

        with open(filepath, "r") as f:
            assert json.loads(f.read()) == {"saucelabs": "default_alias"}

    def test_cannot_rename_cumulusci_default_connected_app(self, keychain):
        error_message = (
            "You cannot rename the connected app service that is provided by CumulusCI."
        )
        with pytest.raises(CumulusCIException, match=error_message):
            keychain.rename_service(
                "connected_app", DEFAULT_CONNECTED_APP_NAME, "new_alias"
            )

    def test_remove_service(self, keychain, service_config):
        home_dir = tempfile.mkdtemp()

        cci_home_dir = Path(f"{home_dir}/.cumulusci")
        cci_home_dir.mkdir()
        with open(cci_home_dir / "DEFAULT_SERVICES.json", "w") as f:
            f.write(json.dumps({"github": "alias"}))

        local_project_dir = cci_home_dir / "test-project"
        local_project_dir.mkdir()
        with open(local_project_dir / "DEFAULT_SERVICES.json", "w") as f:
            f.write(json.dumps({"github": "alias"}))

        github_services_dir = Path(f"{home_dir}/.cumulusci/services/github")
        github_services_dir.mkdir(parents=True)
        self._write_file(github_services_dir / "alias.service", "github config")

        encrypted = keychain._get_config_bytes(service_config)
        keychain.services = {"github": {"alias": encrypted}}
        keychain._default_services["github"] = "alias"
        with mock.patch.object(
            EncryptedFileProjectKeychain, "global_config_dir", cci_home_dir
        ):
            keychain.remove_service("github", "alias")

        # loaded service is removed
        assert "old-alias" not in keychain.services["github"]
        # corresponding .service file is gone
        assert not (github_services_dir / "alias.service").is_file()
        # references in DEFAULT_SERVICES.json are gone
        with open(cci_home_dir / "DEFAULT_SERVICES.json", "r") as f:
            assert json.loads(f.read()) == {}
        with open(local_project_dir / "DEFAULT_SERVICES.json", "r") as f:
            assert json.loads(f.read()) == {}
        # default service is unset
        assert "github" not in keychain._default_services

    def test_remove_service__other_set_as_default(self, keychain, service_config):
        home_dir = tempfile.mkdtemp()

        cci_home_dir = Path(f"{home_dir}/.cumulusci")
        cci_home_dir.mkdir()
        with open(cci_home_dir / "DEFAULT_SERVICES.json", "w") as f:
            f.write(json.dumps({"github": "alias"}))

        local_project_dir = cci_home_dir / "test-project"
        local_project_dir.mkdir()
        with open(local_project_dir / "DEFAULT_SERVICES.json", "w") as f:
            f.write(json.dumps({"github": "alias"}))

        github_services_dir = Path(f"{home_dir}/.cumulusci/services/github")
        github_services_dir.mkdir(parents=True)
        self._write_file(github_services_dir / "alias.service", "github config")

        encrypted = keychain._get_config_bytes(service_config)
        keychain.services = {"github": {"alias": encrypted, "other-service": encrypted}}
        keychain._default_services["github"] = "alias"
        with mock.patch.object(
            EncryptedFileProjectKeychain, "global_config_dir", cci_home_dir
        ):
            keychain.remove_service("github", "alias")

        # the one other service should be the default
        assert keychain._default_services["github"] == "other-service"

    def test_remove_service__cannot_remove_default_connected_app(self, keychain):
        error_message = (
            f"Unable to remove connected app service: {DEFAULT_CONNECTED_APP_NAME}. "
            "This connected app is provided by CumulusCI and cannot be removed."
        )
        with pytest.raises(CumulusCIException, match=error_message):
            keychain.remove_service("connected_app", DEFAULT_CONNECTED_APP_NAME)

    def test_load_default_connected_app(self, keychain):
        keychain._load_default_connected_app()
        assert (
            DEFAULT_CONNECTED_APP_NAME in keychain.config["services"]["connected_app"]
        )
        assert (
            keychain.config["services"]["connected_app"][DEFAULT_CONNECTED_APP_NAME]
            == DEFAULT_CONNECTED_APP
        )

    def test_default_connected_app_should_be_default_after_loading(self, keychain):
        keychain._load_default_connected_app()
        assert keychain._default_services["connected_app"] == DEFAULT_CONNECTED_APP_NAME

    def test_load_default_services(self, keychain):
        expected_defaults = {
            "github": "github_alias",
            "metaci": "metaci_alias",
            "devhub": "devhub_alias",
            "connected_app": "connected_app__alias",
        }
        filepath = Path(f"{keychain.global_config_dir}/DEFAULT_SERVICES.json")
        with open(filepath, "w") as f:
            f.write(json.dumps(expected_defaults))

        keychain._load_default_services()
        assert keychain._default_services["connected_app"] != DEFAULT_CONNECTED_APP_NAME
        assert keychain._default_services == expected_defaults

    def test_read_default_services(self, keychain):
        expected_defaults = {
            "github": "github_alias",
            "metaci": "metaci_alias",
            "devhub": "devhub_alias",
            "connected_app": "connected_app__alias",
        }
        filepath = Path(f"{keychain.global_config_dir}/DEFAULT_SERVICES.json")
        with open(filepath, "w") as f:
            f.write(json.dumps(expected_defaults))

        actual_defaults = keychain._read_default_services(filepath)
        assert actual_defaults == expected_defaults

    def test_read_default_services__file_does_not_exist(self, keychain):
        assert keychain._read_default_services(Path("not-a-valid-filepath")) == {}

    def test_write_default_services(self, keychain):
        expected_defaults = {
            "github": "github_alias",
            "apextestdb": "apextestdb_alias",
            "metaci": "metaci_alias",
            "devhub": "devhub_alias",
            "connected_app": "connected_app__alias",
        }
        filepath = Path(f"{keychain.global_config_dir}/DEFAULT_SERVICES.json")
        keychain._write_default_services(filepath, expected_defaults)

        with open(filepath, "r") as f:
            actual_defaults = json.loads(f.read())
        assert actual_defaults == expected_defaults

    def test_write_default_services__bad_filename(self, keychain):
        filepath = Path("DEFAULT_THINGS.json")
        with pytest.raises(CumulusCIException):
            keychain._write_default_services(filepath, {})

    def test_decrypt_config__no_config(self, keychain):
        config = keychain._decrypt_config(OrgConfig, None, extra=["test", keychain])
        assert config.__class__ == OrgConfig
        assert config.config == {}
        assert config.keychain == keychain

    def test_decrypt_config__no_config_2(self, keychain):
        config = keychain._decrypt_config(BaseConfig, None)
        assert config.__class__ == BaseConfig
        assert config.config == {}

    def test_decrypt_config__Python_2_warning(self, keychain, caplog):
        config = keychain.cleanup_Python_2_configs({"a": "b"})
        assert config == {"a": "b"}
        assert len(caplog.records) == 0

        config = keychain.cleanup_Python_2_configs({"a": b"b", b"c": "d"})
        assert config == {"a": "b", "c": "d"}
        assert len(caplog.records) == 2

    def test_decrypt_config__wrong_key(self, keychain, org_config):
        keychain.set_org(org_config, False)
        keychain.key = "x" * 16
        with pytest.raises(KeychainKeyNotFound):
            keychain.get_org("test")

    def test_validate_key__wrong_length(self, project_config):
        with pytest.raises(ConfigError):
            EncryptedFileProjectKeychain(project_config, "1")

    def test_validate_key__no_key(self, project_config):
        keychain = EncryptedFileProjectKeychain(project_config, None)
        assert keychain._validate_key() is None

    def test_construct_config(self, keychain):
        result = keychain._construct_config(
            None, [{"scratch": "scratch org"}, "org_name"]
        )
        assert isinstance(result, ScratchOrgConfig)
        result = keychain._construct_config(None, [{"sfdx": True}, "org_name"])
        assert isinstance(result, SfdxOrgConfig)

    def test_new_service_type_creates_expected_directory(
        self, keychain, service_config
    ):
        home_dir = tempfile.mkdtemp()
        cci_home_dir = Path(f"{home_dir}/.cumulusci")
        cci_home_dir.mkdir()
        services_dir = cci_home_dir / "services"
        services_dir.mkdir()

        encrypted = keychain._get_config_bytes(service_config)
        with mock.patch.object(
            EncryptedFileProjectKeychain, "global_config_dir", cci_home_dir
        ):
            keychain._save_encrypted_service("new_service_type", "alias", encrypted)

        assert (services_dir / "new_service_type").is_dir()

    @pytest.mark.parametrize(
        "val, expected",
        (
            ("CUMULUSCI_SERVICE_github", ("github", "env")),
            ("CUMULUSCI_SERVICE_github__alias", ("github", "env-alias")),
            ("CUMULUSCI_SERVICE_connected_app", ("connected_app", "env")),
            ("CUMULUSCI_SERVICE_connected_app__alias", ("connected_app", "env-alias")),
        ),
    )
    def test_get_env_service_type_and_name(self, val, expected, project_config):
        keychain = EncryptedFileProjectKeychain(project_config, "0123456789abcdef")
        actual_type, actual_name = keychain._get_env_service_type_and_name(val)
        assert (actual_type, actual_name) == expected

    def test_backwards_compatability_with_EnvironmentProjectKeychain(
        self, project_config, key
    ):
        """Ensure we don't break backwards compatability for people still using EnvironmentProjectKeychain"""
        from cumulusci.core.keychain.environment_project_keychain import (
            EnvironmentProjectKeychain,
        )

        assert EnvironmentProjectKeychain is EncryptedFileProjectKeychain


def _touch_test_org_file(directory):
    org_dir = directory / "orginfo/something.something.saleforce.com"
    org_dir.mkdir(parents=True)
    (org_dir / "testfile.json").touch()
    return org_dir


class TestCleanupOrgCacheDir:
    def test_cleanup_cache_dir(self, keychain):
        keychain.set_org(
            OrgConfig({"instance_url": "http://foo.my.salesforce.com/"}, "dev"), False
        )
        keychain.set_org(
            OrgConfig({"instance_url": "http://bar.my.salesforce.com/"}, "qa"), False
        )

        temp_for_global = tempfile.mkdtemp()
        with mock.patch.object(
            EncryptedFileProjectKeychain, "global_config_dir", Path(temp_for_global)
        ):
            global_org_dir = _touch_test_org_file(keychain.global_config_dir)
            temp_for_project = tempfile.mkdtemp()
            keychain.project_config = mock.Mock()

            cache_dir = keychain.project_config.cache_dir = Path(temp_for_project)
            project_org_dir = _touch_test_org_file(cache_dir)
            with mock.patch(
                "cumulusci.core.keychain.encrypted_file_project_keychain.rmtree"
            ) as rmtree:
                keychain.cleanup_org_cache_dirs()
                rmtree.assert_has_calls(
                    [mock.call(global_org_dir), mock.call(project_org_dir)],
                    any_order=True,
                )

    def test_cleanup_cache_dir__no_project_config(self, keychain):
        keychain.project_config = None
        with mock.patch(
            "cumulusci.core.keychain.encrypted_file_project_keychain.rmtree"
        ) as rmtree:
            keychain.cleanup_org_cache_dirs()
            assert not rmtree.mock_calls, rmtree.mock_calls

    def test_cleanup_cache_dir_nothing_to_cleanup(self, keychain):
        keychain.set_org(
            OrgConfig({"instance_url": "http://foo.my.salesforce.com/"}, "dev"), False
        )

        keychain.project_config = mock.Mock()
        temp_for_global = tempfile.mkdtemp()
        with mock.patch.object(
            EncryptedFileProjectKeychain, "global_config_dir", Path(temp_for_global)
        ):
            temp_for_project = tempfile.mkdtemp()
            cache_dir = keychain.project_config.cache_dir = Path(temp_for_project)
            org_dir = cache_dir / "orginfo/foo.my.salesforce.com"
            org_dir.mkdir(parents=True)
            (org_dir / "schema.json").touch()
            with mock.patch(
                "cumulusci.core.keychain.encrypted_file_project_keychain.rmtree"
            ) as rmtree:
                keychain.cleanup_org_cache_dirs()
                assert not rmtree.mock_calls, rmtree.mock_calls

    duration = (
        (59, "59s"),
        (70, "1m:10s"),
        (119, "1m:59s"),
        (65, "1m:5s"),
        (4000, "1h:6m:40s"),
        (7199, "1h:59m:59s"),
    )

    @pytest.mark.parametrize("val,expected", duration)
    def test_time_delta(self, val, expected):
        formatted = utils.format_duration(datetime.timedelta(seconds=val))
        assert formatted == expected, (formatted, expected)

    def test_set_and_get_org_with_dates__json(
        self, keychain, org_config, key, withdifferentformats
    ):
        org_config.global_org = True
        keychain.key = key

        custom_datetime = datetime.datetime.now()
        custom_date = datetime.datetime.now().date()
        org_config.config["custom_datetime"] = custom_datetime
        org_config.config["custom_date"] = custom_date

        keychain.set_org(org_config, True)
        assert list(keychain.orgs.keys()) == ["test"]
        config = keychain.get_org("test").config
        assert config == {
            **org_config.config,
            "serialization_format": withdifferentformats,
        }
        assert config["custom_datetime"] == custom_datetime
        assert config["custom_date"] == custom_date

    @mock.patch("cumulusci.core.keychain.serialization.SHOULD_SAVE_AS_JSON", True)
    def test_set_and_get_org_with_bad_datatypes(self, keychain, org_config, key):
        org_config.global_org = True
        keychain.key = None
        with mock.patch("pickle.dumps") as dumps:
            dumps.return_value = b"xx"
            org_config.config["good"] = 25

            keychain.set_org(org_config, True)
            assert not dumps.mock_calls

            org_config.config["bad"] = 25j

            keychain.set_org(org_config, True)
            dumps.assert_called_once_with(
                {"foo": "bar", "good": 25, "bad": 25j}, protocol=mock.ANY
            )

    def test_set_and_get_service_with_dates__global(
        self, keychain, key, withdifferentformats
    ):
        service_config = ServiceConfig(
            {"name": "foo1", "password": "1234", "token": "1234"}
        )

        keychain.key = key

        custom_datetime = datetime.datetime.now()
        custom_date = datetime.datetime.now().date()
        service_config.config["custom_datetime"] = custom_datetime
        service_config.config["custom_date"] = custom_date

        keychain.set_service("github", "alias", service_config)
        config = keychain.get_service("github").config
        assert config == {
            **service_config.config,
            "serialization_format": withdifferentformats,
        }
        assert config["custom_datetime"] == custom_datetime
        assert config["custom_date"] == custom_date

    def test_migration_pickle_to_json(
        self, keychain, patch_home_and_env, project_config
    ):
        # This all runs in fake-home, not real home.
        with mock.patch(
            "cumulusci.core.keychain.serialization.SHOULD_SAVE_AS_JSON", False
        ):
            org_config = OrgConfig(
                {"password": "Kltpzyxm"}, "test_migration", keychain=keychain
            )
            org_config.save()
            assert Path(keychain.project_local_dir, "test_migration.org").exists()
            del keychain.config["orgs"]
            keychain._load_orgs()
            assert keychain.get_org("test_migration").password == "Kltpzyxm"
            assert keychain.get_org("test_migration").serialization_format == "pickle"

        with mock.patch(
            "cumulusci.core.keychain.serialization.SHOULD_SAVE_AS_JSON", True
        ):
            org_config.save()
            assert Path(keychain.project_local_dir, "test_migration.org").exists()
            del keychain.config["orgs"]
            keychain._load_orgs()
            assert keychain.get_org("test_migration").password == "Kltpzyxm"
            assert keychain.get_org("test_migration").serialization_format == "json"



from cumulusci.salesforce_api.metadata import ApiRetrieveUnpackaged
from cumulusci.tasks.salesforce import BaseRetrieveMetadata

retrieve_unpackaged_options = BaseRetrieveMetadata.task_options.copy()
retrieve_unpackaged_options.update(
    {
        "package_xml": {
            "description": "The path to a package.xml manifest to use for the retrieve.",
            "required": True,
        },
        "api_version": {
            "description": (
                "Override the default api version for the retrieve."
                + " Defaults to project__package__api_version"
            )
        },
    }
)


class RetrieveUnpackaged(BaseRetrieveMetadata):
    api_class = ApiRetrieveUnpackaged

    task_options = retrieve_unpackaged_options

    def _init_options(self, kwargs):
        super(RetrieveUnpackaged, self)._init_options(kwargs)

        if "package_xml" in self.options:
            self.options["package_xml_path"] = self.options["package_xml"]
            with open(self.options["package_xml_path"], "r") as f:
                self.options["package_xml"] = f.read()

    def _get_api(self):
        return self.api_class(
            self, self.options["package_xml"], self.options.get("api_version")
        )



import datetime
import json
import os
from unittest import mock

import pytest
import responses

from cumulusci.core.exceptions import CumulusCIException
from cumulusci.tasks.salesforce import DescribeMetadataTypes
from cumulusci.tasks.salesforce.nonsourcetracking import (
    ListComponents,
    ListNonSourceTrackable,
    RetrieveComponents,
)
from cumulusci.tests.util import create_project_config
from cumulusci.utils import temporary_dir


class TestListNonSourceTrackable:
    @pytest.mark.parametrize(
        "json_data, expected_status,error_msg",
        [
            (None, 404, "Failed to retrieve response with status code 404"),
            (
                {"types": {"Scontrol": {"channels": {}}}},
                200,
                "Api version 44.0 not supported",
            ),
        ],
    )
    @responses.activate
    def test_run_get_types_details(
        self, create_task_fixture, json_data, expected_status, error_msg
    ):
        options = {"api_version": 44.0}
        responses.add(
            "GET",
            f"https://dx-extended-coverage.my.salesforce-sites.com/services/apexrest/report?version={options['api_version']}",
            json=json_data,
            status=expected_status,
        )
        task = create_task_fixture(ListNonSourceTrackable, options)
        task._init_task()
        with pytest.raises(CumulusCIException, match=error_msg):
            task._run_task()

    @responses.activate
    def test_run_task_list_metadatatypes(self, create_task_fixture):
        options = {"api_version": 44.0}
        return_value = {
            "types": {
                "Scontrol": {
                    "channels": {"sourceTracking": False, "metadataApi": True}
                },
                "SharingRules": {
                    "channels": {"sourceTracking": False, "metadataApi": True}
                },
                "ApexClass": {
                    "channels": {"sourceTracking": True, "metadataApi": True}
                },
                "CustomObject": {
                    "channels": {"sourceTracking": True, "metadataApi": True}
                },
                "MessagingChannel": {
                    "channels": {"sourceTracking": True, "metadataApi": True}
                },
            }
        }
        responses.add(
            "GET",
            f"https://dx-extended-coverage.my.salesforce-sites.com/services/apexrest/report?version={options['api_version']}",
            json=return_value,
            status=200,
        )

        task = create_task_fixture(ListNonSourceTrackable, options)
        task._init_task()
        with mock.patch.object(
            DescribeMetadataTypes,
            "_run_task",
            return_value=["Scontrol", "ApexClass", "CustomObject", "SharingRules"],
        ):
            non_source_trackable = task._run_task()
            assert non_source_trackable == ["Scontrol", "SharingRules"]


class TestListComponents:
    def test_init_task(self, create_task_fixture):
        with mock.patch.object(
            ListNonSourceTrackable,
            "_run_task",
            return_value=["SharingRules", "Scontrol"],
        ):
            task = create_task_fixture(ListComponents, {})
            task._init_task()
            assert task.options["metadata_types"] == ["SharingRules", "Scontrol"]

    def test_check_api_result(self, create_task_fixture):
        options = {"metadata_types": "SharingRules"}
        task = create_task_fixture(ListComponents, options)
        expected = [
            {
                "MemberType": "SharingRules",
                "MemberName": "Order",
                "lastModifiedByName": "Automated process",
                "lastModifiedDate": datetime.datetime(1970, 1, 1, 0, 0),
            },
            {
                "MemberType": "SharingRules",
                "MemberName": "BusinessBrand",
                "lastModifiedByName": "User User",
                "lastModifiedDate": datetime.datetime(1970, 1, 1, 0, 0),
            },
        ]
        result = mock.Mock(
            return_value={
                "SharingRules": [
                    {
                        "createdById": "0055j000008HpiJAAS",
                        "createdByName": "Alpha",
                        "createdDate": datetime.datetime(1970, 1, 1, 0, 0),
                        "fileName": "sharingRules/Order.sharingRules",
                        "fullName": "Order",
                        "id": None,
                        "lastModifiedById": "0055j000008HpiJAAS",
                        "lastModifiedByName": "Automated process",
                        "lastModifiedDate": datetime.datetime(1970, 1, 1, 0, 0),
                        "manageableState": None,
                        "namespacePrefix": None,
                        "type": "SharingRules",
                    },
                    {
                        "createdById": "0055j000008HpiJAAS",
                        "createdByName": "Beta",
                        "createdDate": datetime.datetime(1970, 1, 1, 0, 0),
                        "fileName": "sharingRules/BusinessBrand.sharingRules",
                        "fullName": "BusinessBrand",
                        "id": None,
                        "lastModifiedById": "0055j000008HpiJAAS",
                        "lastModifiedByName": "User User",
                        "lastModifiedDate": datetime.datetime(1970, 1, 1, 0, 0),
                        "manageableState": None,
                        "namespacePrefix": None,
                        "type": "SharingRules",
                    },
                ]
            }
        )
        messages = []
        task._init_task()
        task.logger = mock.Mock()
        task.logger.info = messages.append
        task.api_class = mock.Mock(return_value=result)
        components = task._run_task()
        assert components == expected
        assert (
            "Found 2 non source trackable components in the org for the given types."
            in messages
        )


@mock.patch("cumulusci.tasks.salesforce.sourcetracking.sfdx")
class TestRetrieveComponents:
    def test_init_options__sfdx_format(self, sfdx, create_task_fixture):
        with temporary_dir():
            project_config = create_project_config()
            project_config.project__source_format = "sfdx"
            with open("sfdx-project.json", "w") as f:
                json.dump(
                    {"packageDirectories": [{"path": "force-app", "default": True}]}, f
                )
            task = create_task_fixture(
                RetrieveComponents, {"metadata_types": "SharingRules"}, project_config
            )
            assert not task.md_format
            assert task.options["path"] == "force-app"

    def test_run_task(self, sfdx, create_task_fixture):
        sfdx_calls = []
        sfdx.side_effect = lambda cmd, *args, **kw: sfdx_calls.append(cmd)

        with temporary_dir():
            task = create_task_fixture(
                RetrieveComponents,
                {
                    "include": "alpha",
                    "namespace_tokenize": "ns",
                    "metadata_types": "SharingRules",
                },
            )
            task._init_task()
            messages = []
            with mock.patch.object(
                ListComponents,
                "_get_components",
                return_value=[
                    {
                        "MemberType": "SharingRules",
                        "MemberName": "alpha",
                        "lastModifiedByName": "Automated process",
                        "lastModifiedDate": datetime.datetime(1970, 1, 1, 0, 0),
                    },
                    {
                        "MemberType": "SharingRules",
                        "MemberName": "BusinessBrand",
                        "lastModifiedByName": "User User",
                        "lastModifiedDate": datetime.datetime(1970, 1, 1, 0, 0),
                    },
                ],
            ):
                task.logger = mock.Mock()
                task.logger.info = messages.append
                task._run_task()
                assert "SharingRules: alpha" in messages
                assert "SharingRules: BusinessBrand" not in messages
                assert sfdx_calls == [
                    "force:mdapi:convert",
                    "force:source:retrieve",
                    "force:source:convert",
                ]
                assert os.path.exists(os.path.join("src", "package.xml"))

    def test_run_task__no_changes(self, sfdx, create_task_fixture):
        with temporary_dir() as path:
            task = create_task_fixture(
                RetrieveComponents, {"path": path, "metadata_types": "SharingRules"}
            )
            task._init_task()
            messages = []
            with mock.patch.object(ListComponents, "_get_components", return_value=[]):
                task.logger = mock.Mock()
                task.logger.info = messages.append
                task._run_task()
                assert "No components to retrieve" in messages



import os
import time
from datetime import datetime

import pytz

from cumulusci.core.github import get_tag_by_name
from cumulusci.utils.version_strings import LooseVersion


class BaseChangeNotesProvider(object):
    def __init__(self, release_notes_generator):
        self.release_notes_generator = release_notes_generator

    def __call__(self):
        """Subclasses should provide an implementation that returns an
        iterable of each change note"""
        raise NotImplementedError()


class StaticChangeNotesProvider(BaseChangeNotesProvider):
    def __init__(self, release_notes_generator, change_notes):
        super(StaticChangeNotesProvider, self).__init__(release_notes_generator)
        self.change_notes = change_notes

    def __call__(self):
        for change_note in self.change_notes:
            yield change_note


class DirectoryChangeNotesProvider(BaseChangeNotesProvider):
    def __init__(self, release_notes_generator, directory):
        super(DirectoryChangeNotesProvider, self).__init__(release_notes_generator)
        self.directory = directory

    def __call__(self):
        for item in sorted(os.listdir(self.directory)):
            yield open("{}/{}".format(self.directory, item)).read()


class GithubChangeNotesProvider(BaseChangeNotesProvider):
    """Provides changes notes by finding all merged pull requests to
    the default branch between two tags.

    Expects the passed release_notes_generator instance to have a github_info
    property that contains a dictionary of settings for accessing Github:

        - github_repo
        - github_owner
        - github_username
        - github_password
        - default_branch
        - prefix_prod: Tag prefix for production release tags. Defaults to 'prod/'
    """

    def __init__(self, release_notes_generator, current_tag, last_tag=None):
        super(GithubChangeNotesProvider, self).__init__(release_notes_generator)
        self.current_tag = current_tag
        self._last_tag = last_tag
        self._start_date = None
        self._end_date = None
        self.repo = release_notes_generator.get_repo()
        self.github_info = release_notes_generator.github_info

    def __call__(self):
        for pull_request in self._get_pull_requests():
            yield pull_request

    @property
    def last_tag(self):
        if not self._last_tag:
            self._last_tag = self._get_last_tag()
        return self._last_tag

    @property
    def current_tag_info(self):
        if not hasattr(self, "_current_tag_info"):
            tag = get_tag_by_name(self.repo, self.current_tag)
            self._current_tag_info = {"tag": tag, "commit": self._get_commit_info(tag)}
        return self._current_tag_info

    @property
    def last_tag_info(self):
        if not hasattr(self, "_last_tag_info"):
            if self.last_tag:
                tag = get_tag_by_name(self.repo, self.last_tag)
                self._last_tag_info = {"tag": tag, "commit": self._get_commit_info(tag)}
            else:
                self._last_tag_info = None
        return self._last_tag_info

    def _get_commit_info(self, tag):
        return self.repo.git_commit(tag.object.sha)

    @property
    def start_date(self):
        return self._get_commit_date(self.current_tag_info["commit"])

    @property
    def end_date(self):
        if self.last_tag_info:
            return self._get_commit_date(self.last_tag_info["commit"])

    def _get_commit_date(self, commit):
        t = time.strptime(commit.author["date"], "%Y-%m-%dT%H:%M:%SZ")
        return datetime(t[0], t[1], t[2], t[3], t[4], t[5], t[6], pytz.UTC)

    def _get_version_from_tag(self, tag):
        if tag.startswith(self.github_info["prefix_prod"]):
            return tag.replace(self.github_info["prefix_prod"], "")
        elif tag.startswith(self.github_info["prefix_beta"]):
            return tag.replace(self.github_info["prefix_beta"], "")
        raise ValueError("Could not determine version number from tag {}".format(tag))

    def _get_last_tag(self):
        """Gets the last release tag before self.current_tag"""

        current_version = LooseVersion(
            self._get_version_from_tag(self.release_notes_generator.current_tag)
        )

        versions = []
        for tag in self.repo.tags():
            if not tag.name.startswith(self.github_info["prefix_prod"]):
                continue
            version = LooseVersion(self._get_version_from_tag(tag.name))
            if version >= current_version:
                continue
            versions.append(version)
        if versions:
            versions.sort()
            return "{}{}".format(self.github_info["prefix_prod"], versions[-1])

    def _get_pull_requests(self):
        """Gets all pull requests from the repo since we can't do a filtered
        date merged search"""
        for pull in self.repo.pull_requests(
            state="closed", base=self.github_info["default_branch"], direction="asc"
        ):
            if self._include_pull_request(pull):
                yield pull

    def _include_pull_request(self, pull_request):
        """Checks if the given pull_request was merged to the default branch
        between self.start_date and self.end_date"""

        merged_date = pull_request.merged_at
        if not merged_date:
            return False
        if self.last_tag:
            last_tag_sha = self.last_tag_info["commit"].sha
            if pull_request.merge_commit_sha == last_tag_sha:
                # Github commit dates can be different from the merged_at date
                return False

        current_tag_sha = self.current_tag_info["commit"].sha
        if pull_request.merge_commit_sha == current_tag_sha:
            return True

        # include PRs before current tag
        if merged_date <= self.start_date:
            if self.end_date:
                # include PRs after last tag
                if (
                    merged_date > self.end_date
                    and pull_request.merge_commit_sha != last_tag_sha
                ):
                    return True
            else:
                # no last tag, include all PRs before current tag
                return True

        return False



import json
import os
import pathlib
from unittest import mock

import pytest

from cumulusci.core.config import OrgConfig
from cumulusci.core.exceptions import ProjectConfigNotFound
from cumulusci.tasks.salesforce.retrieve_profile import RetrieveProfile
from cumulusci.tasks.salesforce.sourcetracking import (
    KNOWN_BAD_MD_TYPES,
    ListChanges,
    RetrieveChanges,
    SnapshotChanges,
    _write_manifest,
    retrieve_components,
)
from cumulusci.tests.util import create_project_config
from cumulusci.utils import temporary_dir


class TestListChanges:
    """List the changes from a scratch org"""

    def test_run_task(self, create_task_fixture):
        task = create_task_fixture(ListChanges, {"exclude": "Ignore"})
        task._init_task()
        task.tooling = mock.Mock()
        task.logger = mock.Mock()
        task.tooling.query_all.return_value = {
            "totalSize": 1,
            "records": [
                {
                    "MemberType": "CustomObject",
                    "MemberName": "Test__c",
                    "RevisionCounter": 1,
                },
                {
                    "MemberType": "CustomObject",
                    "MemberName": "Ignored__c",
                    "RevisionCounter": 2,
                },
            ],
        }
        task._run_task()
        assert "CustomObject: Test__c" in task.logger.info.call_args[0][0]

    def test_run_task__no_changes(self, create_task_fixture):
        task = create_task_fixture(ListChanges)
        task._init_task()
        task.tooling = mock.Mock()
        task.logger = mock.Mock()
        task.tooling.query_all.return_value = {"totalSize": 0, "records": []}
        task._run_task()
        assert "Found no changes." in task.logger.info.call_args[0][0]

    def test_run_task__snapshot(self, create_task_fixture):
        with temporary_dir():
            task = create_task_fixture(ListChanges, {"snapshot": True})
            task._init_task()
            task.tooling = mock.Mock()
            messages = []
            task.logger = mock.Mock()
            task.logger.info = messages.append
            task.tooling.query_all.return_value = {
                "totalSize": 1,
                "records": [
                    {
                        "MemberType": "CustomObject",
                        "MemberName": "Test__c",
                        "RevisionCounter": 1,
                    }
                ],
            }
            task._run_task()
            assert os.path.exists(
                os.path.join(task.project_config.cache_dir, "snapshot", "test.json")
            )

            assert "CustomObject: Test__c" in messages

            task = create_task_fixture(ListChanges)
            task._init_task()
            task.tooling = mock.Mock()
            task.logger = mock.Mock()
            task.logger.info = messages.append
            task.tooling.query_all.return_value = {
                "totalSize": 1,
                "records": [
                    {
                        "MemberType": "CustomObject",
                        "MemberName": "Test__c",
                        "RevisionCounter": 1,
                    }
                ],
            }
            task._run_task()
            assert "Found no changes." in messages

    def test_filter_changes__include(self, create_task_fixture):
        foo = {
            "MemberType": "CustomObject",
            "MemberName": "foo__c",
            "RevisionCounter": 1,
        }
        bar = {
            "MemberType": "CustomObject",
            "MemberName": "bar__c",
            "RevisionCounter": 1,
        }
        foobar = {
            "MemberType": "CustomObject",
            "MemberName": "foobar__c",
            "RevisionCounter": 1,
        }
        task = create_task_fixture(ListChanges, {"include": "foo", "exclude": "bar"})
        filtered, ignored = task._filter_changes([foo, bar, foobar])
        assert filtered == [foo]

    def test_filter_changes__null_revnum(self, create_task_fixture):
        foo = {
            "MemberType": "CustomObject",
            "MemberName": "foo__c",
            "RevisionCounter": None,
        }
        bar = {
            "MemberType": "CustomObject",
            "MemberName": "bar__c",
            "RevisionCounter": 1,
        }
        task = create_task_fixture(ListChanges, {})
        filtered, ignored = task._filter_changes([foo, bar])
        assert filtered == [foo, bar]


@mock.patch("cumulusci.tasks.salesforce.sourcetracking.sfdx")
class TestRetrieveChanges:
    """Retrieve changed components from a scratch org"""

    def test_init_options__sfdx_format(self, sfdx, create_task_fixture):
        with temporary_dir():
            project_config = create_project_config()
            project_config.project__source_format = "sfdx"
            with open("sfdx-project.json", "w") as f:
                json.dump(
                    {"packageDirectories": [{"path": "force-app", "default": True}]}, f
                )
            task = create_task_fixture(RetrieveChanges, {}, project_config)
            assert not task.md_format
            assert task.options["path"] == "force-app"

    def test_run_task(self, sfdx, create_task_fixture):
        sfdx_calls = []
        sfdx.side_effect = lambda cmd, *args, **kw: sfdx_calls.append(cmd)

        with temporary_dir():
            task = create_task_fixture(
                RetrieveChanges,
                {
                    "include": "Test",
                    "namespace_tokenize": "ns",
                    "retrieve_complete_profile": True,
                },
            )
            task._init_task()
            task.tooling = mock.Mock()
            task.tooling.query_all.return_value = {
                "totalSize": 1,
                "records": [
                    {
                        "MemberType": "CustomObject",
                        "MemberName": "Test__c",
                        "RevisionCounter": 1,
                    },
                    {
                        "MemberType": "Profile",
                        "MemberName": "TestProfile",
                        "RevisionCounter": 1,
                    },
                ],
            }
            with mock.patch.object(
                RetrieveProfile, "_run_task"
            ) as mock_retrieve_profile, mock.patch.object(
                pathlib.Path, "exists", return_value=True
            ), mock.patch.object(
                pathlib.Path, "is_dir", return_value=True
            ):
                task._run_task()

                assert sfdx_calls == [
                    "force:mdapi:convert",
                    "force:source:retrieve",
                    "force:source:convert",
                ]
                assert os.path.exists(os.path.join("src", "package.xml"))
                mock_retrieve_profile.assert_called()

    def test_run_task__no_changes(self, sfdx, create_task_fixture):
        with temporary_dir() as path:
            task = create_task_fixture(RetrieveChanges, {"path": path})
            task._init_task()
            messages = []
            task.tooling = mock.Mock()
            task.tooling.query_all.return_value = {"totalSize": 0, "records": []}
            task.logger = mock.Mock()
            task.logger.info = messages.append
            task._run_task()
            assert "No changes to retrieve" in messages


class TestSnapshotChanges:
    @mock.patch("cumulusci.tasks.salesforce.sourcetracking.sfdx")
    def test_run_task(self, sfdx, create_task_fixture):
        with temporary_dir():
            org_config = OrgConfig(
                {
                    "username": "test-cci@example.com",
                    "scratch": True,
                    "instance_url": "https://test.salesforce.com",
                    "access_token": "TOKEN",
                },
                "test",
            )
            task = create_task_fixture(SnapshotChanges, org_config=org_config)
            task._init_task()
            task.tooling.query = mock.Mock(
                side_effect=[
                    {"totalSize": 0, "records": [], "done": True},
                    {
                        "totalSize": 1,
                        "done": True,
                        "records": [
                            {
                                "MemberType": "CustomObject",
                                "MemberName": "Object2",
                                "RevisionCounter": 1,
                            }
                        ],
                    },
                ]
            )
            task._reset_sfdx_snapshot = mock.Mock()
            task._run_task()
            task._reset_sfdx_snapshot.assert_called_once()

    def test_freeze(self, create_task_fixture):
        task = create_task_fixture(SnapshotChanges)
        steps = task.freeze(None)
        assert steps == []


def test_write_manifest__folder():
    with temporary_dir() as path:
        _write_manifest(
            [{"MemberType": "ReportFolder", "MemberName": "TestFolder"}], path, "52.0"
        )
        package_xml = pathlib.Path(path, "package.xml").read_text()
        assert "<name>Report</name>" in package_xml


def test_write_manifest__bad_md_types():
    benign_md_type = [{"MemberType": "ReportFolder", "MemberName": "TestFolder"}]
    bad_md_types = [
        {"MemberType": type_name, "MemberName": "Banana"}
        for type_name in KNOWN_BAD_MD_TYPES
    ]
    with temporary_dir() as path:
        changes = benign_md_type + bad_md_types
        _write_manifest(changes, path, "52.0")
        package_xml = pathlib.Path(path, "package.xml").read_text()
        assert "<name>Report</name>" in package_xml
        for name in bad_md_types:
            assert f"<name>{name}</name>" not in package_xml


def test_retrieve_components_project_config_not_found():
    components = mock.Mock()
    org_config = mock.Mock()
    target = "force-app/"
    md_format = False
    extra_package_xml_opts = {"sample": "dict"}
    namespace_tokenize = "sample"
    api_version = "58.0"
    expected_error_message = (
        "Kindly provide project_config as part of retrieve_components"
    )
    with pytest.raises(ProjectConfigNotFound) as e:
        retrieve_components(
            components=components,
            org_config=org_config,
            target=target,
            md_format=md_format,
            extra_package_xml_opts=extra_package_xml_opts,
            namespace_tokenize=namespace_tokenize,
            api_version=api_version,
            retrieve_complete_profile=True,
        )
        assert expected_error_message == e.value.message



"""
This is a library used by locators.robot for testing
custom locator strategies
"""
from cumulusci.robotframework.locator_manager import (
    register_locators,
    translate_locator,
)

locators = {"appname": "//div[contains(@class, 'appName') and .='{}']"}


class TestLibraryB:
    ROBOT_LIBRARY_SCOPE = "global"

    def __init__(self):
        register_locators("B", locators)

    def translate_locator(self, prefix, locator):
        return translate_locator(prefix, locator)



import json
from inspect import signature
from typing import Any, Dict, List

from pydantic import DirectoryPath, Field, FilePath, create_model

from cumulusci.core.exceptions import TaskOptionsError
from cumulusci.utils.yaml.model_parser import CCIDictModel

READONLYDICT_ERROR_MSG = (
    "The 'options' dictionary is read-only. Please use 'parsed_options' instead."
)


def _describe_field(field):
    "Convert a Pydantic field into a CCI task_option dict"
    rc = {
        "description": field.field_info.description,
        "required": field.required,
    }
    if field.field_info.default != ...:
        rc["default"] = field.field_info.default
    return rc


class ReadOnlyOptions(dict):
    """To enforce self.options to be read-only"""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def __setitem__(self, key, value):
        raise TaskOptionsError(READONLYDICT_ERROR_MSG)

    def __delitem__(self, key):
        raise TaskOptionsError(READONLYDICT_ERROR_MSG)

    def pop(self, key, default=None):
        raise TaskOptionsError(READONLYDICT_ERROR_MSG)


class CCIOptions(CCIDictModel):
    "Base class for all options in tasks"

    @classmethod
    def as_task_options(cls):
        return {
            fieldname: _describe_field(field)
            for fieldname, field in cls.__fields__.items()
        }


class CCIOptionType:
    """Base class for custom option types.

    Subclasses must implement 'from_str' and it must have a type-hinted return value like this:

    class ListOfStringsOption(CCIOptionType):
        @classmethod
        def from_str(cls, v) -> List[str]:
            return [s.strip() for s in v.split(",")]
    """

    name = None

    @classmethod
    def __get_validators__(cls):
        "https://pydantic-docs.helpmanual.io/usage/types/#classes-with-__get_validators__"
        yield cls.validate

    @classmethod
    def validate(cls, v):
        """Validate and convert a value.

        If its a string, parse it, else, just validate it.
        """
        if isinstance(v, str):
            if v.startswith("{") or v.startswith("["):
                v = json.loads(v)
            else:
                v = cls.from_str(v)

        # Pydantic can't just parse/validate arbitrary data unless
        # it has a model. So we create a dummy model for
        # it to have a parsing/validating context
        target_type = signature(cls.from_str).return_annotation

        Dummy = create_model(cls.name or cls.__name__, __root__=(target_type, ...))

        return Dummy.parse_obj(v).__root__


class ListOfStringsOption(CCIOptionType):
    """Parses a list of strings from a comma-delimited string"""

    @classmethod
    def from_str(cls, v) -> List[str]:
        return [s.strip() for s in v.split(",")]


class MappingOption(CCIOptionType):
    """Parses a Mapping of Str->Any from a string in format a:b,c:d"""

    @classmethod
    def from_str(cls, v) -> Dict[str, Any]:
        return parse_list_of_pairs_dict_arg(v)


def parse_list_of_pairs_dict_arg(arg):
    """Process an arg in the format "aa:bb,cc:dd" """
    if isinstance(arg, dict):
        return arg
    elif isinstance(arg, str):
        rc = {}
        for key_value in arg.split(","):
            subparts = key_value.split(":", maxsplit=1)
            if len(subparts) == 2:
                key, value = subparts
                if key in rc:
                    raise TypeError(f"Var specified twice: {key}")
                rc[key] = value
            else:
                raise TypeError(f"Var is not a name/value pair: {key_value}")
        return rc
    else:
        raise TypeError(f"Arg is not a dict or string ({type(arg)}): {arg}")


__all__ = [
    "Field",
    "FilePath",
    "DirectoryPath",
    "parse_list_of_pairs_dict_arg",
    "CCIOptions",
    "CCIOptionType",
    "ListOfStringsOption",
    "MappingOption",
]



from collections import defaultdict

from cumulusci.tasks.salesforce import BaseSalesforceApiTask


class CheckSObjectRecordTypes(BaseSalesforceApiTask):
    def _run_task(self):
        rts = defaultdict(list)
        records = self.tooling.query_all(
            "Select SobjectType, FullName FROM RecordType"
        )["records"]
        for r in records:
            rts[r["SobjectType"]].append(r["FullName"].split(".")[1])

        self.return_values = rts
        self.logger.info(f"Found existing Record Types: {self.return_values}")



import abc
import itertools
from typing import Callable, Iterable, List, Optional, Tuple, Union

from github3.exceptions import NotFoundError
from github3.repos.branch import Branch
from github3.repos.commit import RepoCommit, ShortCommit
from github3.repos.repo import Repository

from cumulusci.core.config.project_config import BaseProjectConfig
from cumulusci.core.dependencies.dependencies import (
    BaseGitHubDependency,
    Dependency,
    DependencyPin,
    DynamicDependency,
    PackageNamespaceVersionDependency,
    PackageVersionIdDependency,
    StaticDependency,
    parse_dependencies,
    parse_pins,
)
from cumulusci.core.dependencies.github import (
    get_package_data,
    get_package_details_from_tag,
    get_remote_project_config,
    get_repo,
)
from cumulusci.core.enums import StrEnum
from cumulusci.core.exceptions import CumulusCIException, DependencyResolutionError
from cumulusci.core.github import (
    find_latest_release,
    find_repo_feature_prefix,
    get_version_id_from_commit,
)
from cumulusci.core.versions import PackageType
from cumulusci.utils.git import (
    construct_release_branch_name,
    get_feature_branch_name,
    get_release_identifier,
    is_release_branch_or_child,
)


class DependencyResolutionStrategy(StrEnum):
    """Enum that defines a strategy for resolving a dynamic dependency into a static dependency."""

    STATIC_TAG_REFERENCE = "tag"
    COMMIT_STATUS_EXACT_BRANCH = "commit_status_exact_branch"
    COMMIT_STATUS_RELEASE_BRANCH = "commit_status_release_branch"
    COMMIT_STATUS_PREVIOUS_RELEASE_BRANCH = "commit_status_previous_release_branch"
    COMMIT_STATUS_DEFAULT_BRANCH = "commit_status_default_branch"
    UNLOCKED_EXACT_BRANCH = "unlocked_exact_branch"
    UNLOCKED_RELEASE_BRANCH = "unlocked_release_branch"
    UNLOCKED_PREVIOUS_RELEASE_BRANCH = "unlocked_previous_release_branch"
    UNLOCKED_DEFAULT_BRANCH = "unlocked_default_branch"
    BETA_RELEASE_TAG = "latest_beta"
    RELEASE_TAG = "latest_release"
    UNMANAGED_HEAD = "unmanaged"


class AbstractResolver(abc.ABC):
    """Abstract base class for dependency resolution strategies."""

    name = "Resolver"

    @abc.abstractmethod
    def can_resolve(self, dep: DynamicDependency, context: BaseProjectConfig) -> bool:
        pass

    @abc.abstractmethod
    def resolve(
        self, dep: DynamicDependency, context: BaseProjectConfig
    ) -> Tuple[Optional[str], Optional[StaticDependency]]:
        pass

    def __str__(self):
        return self.name


class GitHubTagResolver(AbstractResolver):
    """Resolver that identifies a ref by a specific GitHub tag."""

    name = "GitHub Tag Resolver"

    def can_resolve(self, dep: DynamicDependency, context: BaseProjectConfig) -> bool:
        return isinstance(dep, BaseGitHubDependency) and dep.tag is not None

    def resolve(
        self, dep: BaseGitHubDependency, context: BaseProjectConfig
    ) -> Tuple[Optional[str], Optional[StaticDependency]]:
        try:
            # Find the github release corresponding to this tag.
            repo = get_repo(dep.github, context)
            release = repo.release_from_tag(dep.tag)
            tag = repo.tag(repo.ref(f"tags/{release.tag_name}").object.sha)
            ref = tag.object.sha
            package_config = get_remote_project_config(repo, ref)
            package_name, namespace = get_package_data(package_config)
            version_id, package_type = get_package_details_from_tag(tag)

            install_unmanaged = (
                dep.is_unmanaged  # We've been told to use this dependency unmanaged
                or not (
                    # We will install managed if:
                    namespace  # the package has a namespace
                    or version_id  # or is a non-namespaced Unlocked Package
                )
            )

            if install_unmanaged:
                return ref, None
            else:
                if package_type is PackageType.SECOND_GEN:
                    package_dep = PackageVersionIdDependency(
                        version_id=version_id,
                        version_number=release.name,
                        package_name=package_name,
                    )
                else:
                    package_dep = PackageNamespaceVersionDependency(
                        namespace=namespace,
                        version=release.name,
                        package_name=package_name,
                        version_id=version_id,
                    )

                return (ref, package_dep)
        except NotFoundError:
            raise DependencyResolutionError(f"No release found for tag {dep.tag}")


class GitHubReleaseTagResolver(AbstractResolver):
    """Resolver that identifies a ref by finding the latest GitHub release."""

    name = "GitHub Release Resolver"
    include_beta = False

    def can_resolve(self, dep: DynamicDependency, context: BaseProjectConfig) -> bool:
        return isinstance(dep, BaseGitHubDependency)

    def resolve(
        self, dep: BaseGitHubDependency, context: BaseProjectConfig
    ) -> Tuple[Optional[str], Optional[StaticDependency]]:
        repo = get_repo(dep.github, context)
        release = find_latest_release(repo, include_beta=self.include_beta)
        if release:
            tag = repo.tag(repo.ref(f"tags/{release.tag_name}").object.sha)
            version_id, package_type = get_package_details_from_tag(tag)

            ref = tag.object.sha
            package_config = get_remote_project_config(repo, ref)
            package_name, namespace = get_package_data(package_config)

            install_unmanaged = (
                dep.is_unmanaged  # We've been told to use this dependency unmanaged
                or not (
                    # We will install managed if:
                    namespace  # the package has a namespace
                    or version_id  # or is a non-namespaced Unlocked Package
                )
            )

            if install_unmanaged:
                return ref, None
            else:
                if package_type is PackageType.SECOND_GEN:
                    package_dep = PackageVersionIdDependency(
                        version_id=version_id,
                        version_number=release.name,
                        package_name=package_name,
                    )
                else:
                    package_dep = PackageNamespaceVersionDependency(
                        namespace=namespace,
                        version=release.name,
                        package_name=package_name,
                        version_id=version_id,
                    )
                return (ref, package_dep)

        return (None, None)


class GitHubBetaReleaseTagResolver(GitHubReleaseTagResolver):
    """Resolver that identifies a ref by finding the latest GitHub release, including betas."""

    name = "GitHub Release Resolver (Betas)"
    include_beta = True


class GitHubUnmanagedHeadResolver(AbstractResolver):
    """Resolver that identifies a ref by finding the latest commit on the main branch."""

    name = "GitHub Unmanaged Resolver"

    def can_resolve(self, dep: DynamicDependency, context: BaseProjectConfig) -> bool:
        return isinstance(dep, BaseGitHubDependency)

    def resolve(
        self, dep: BaseGitHubDependency, context: BaseProjectConfig
    ) -> Tuple[Optional[str], Optional[StaticDependency]]:
        repo = get_repo(dep.github, context)
        return (repo.branch(repo.default_branch).commit.sha, None)


def get_release_id(context: BaseProjectConfig) -> int:
    """Detect a release id (like NNN in feature/NNN__some_branch)
    in the current branch and return it as an integer."""
    if not context.repo_branch or not context.project__git__prefix_feature:
        raise DependencyResolutionError(
            "Cannot get current branch or feature branch prefix"
        )
    release_id = get_release_identifier(
        context.repo_branch, context.project__git__prefix_feature
    )
    if not release_id:
        raise DependencyResolutionError("Cannot get current release identifier")

    return int(release_id)


def locate_commit_status_package_id(
    remote_repo: Repository, release_branch: Branch, context_2gp: str
) -> Tuple[Optional[str], Optional[Union[RepoCommit, ShortCommit]]]:
    """Given a branch on a remote repo, walk the first 5 commits looking
    for a commit status equal to context_2gp and attempt to parse a
    package version id from the commit status detail."""
    version_id = None
    count = 0
    commit = release_branch.commit
    while version_id is None and count < 5:
        version_id = get_version_id_from_commit(remote_repo, commit.sha, context_2gp)
        if version_id:
            break
        count += 1
        if commit.parents:
            commit = remote_repo.commit(commit.parents[0]["sha"])
        else:
            commit = None
            break

    return version_id, commit


def get_remote_context(
    repo: Repository, commit_status_context: str, default_context: str
) -> str:
    config = get_remote_project_config(repo, repo.default_branch)
    return config.lookup(f"project__git__{commit_status_context}") or default_context


class AbstractGitHubCommitStatusPackageResolver(AbstractResolver, abc.ABC):
    """Abstract base class for resolvers that use commit statuses to find packages."""

    commit_status_context = ""
    commit_status_default = ""

    def can_resolve(self, dep: DynamicDependency, context: BaseProjectConfig) -> bool:
        return self.is_valid_repo_context(context) and isinstance(
            dep, BaseGitHubDependency
        )

    def is_valid_repo_context(self, context: BaseProjectConfig) -> bool:
        return bool(context.repo_branch and context.project__git__prefix_feature)

    @abc.abstractmethod
    def get_branches(
        self,
        dep: BaseGitHubDependency,
        context: BaseProjectConfig,
    ) -> List[Branch]:
        ...

    def resolve(
        self, dep: BaseGitHubDependency, context: BaseProjectConfig
    ) -> Tuple[Optional[str], Optional[StaticDependency]]:
        branches = self.get_branches(dep, context)

        # We know `repo` is not None because `get_branches()` will raise in that case.
        repo = context.get_repo_from_url(dep.github)
        remote_context = get_remote_context(
            repo, self.commit_status_context, self.commit_status_default
        )
        for branch in branches:
            version_id, commit = locate_commit_status_package_id(
                repo,
                branch,
                remote_context,
            )

            if version_id and commit:
                context.logger.info(
                    f"{self.name} located package version {version_id} on branch {branch.name} on {repo.clone_url} at commit {branch.commit.sha}"
                )
                package_config = get_remote_project_config(repo, commit.sha)
                package_name, _ = get_package_data(package_config)

                return commit.sha, PackageVersionIdDependency(
                    version_id=version_id, package_name=package_name
                )

        context.logger.warn(
            f"{self.name} did not locate package package version on {repo.clone_url}."
        )
        return (None, None)


class AbstractGitHubReleaseBranchResolver(
    AbstractGitHubCommitStatusPackageResolver, abc.ABC
):
    """Abstract base class for resolvers that use commit statuses on release branches to find refs."""

    branch_offset_start = 0
    branch_offset_end = 0

    def is_valid_repo_context(self, context: BaseProjectConfig) -> bool:
        return bool(
            super().is_valid_repo_context(context)
            and is_release_branch_or_child(
                context.repo_branch, context.project__git__prefix_feature  # type: ignore
            )
        )

    def get_branches(
        self,
        dep: BaseGitHubDependency,
        context: BaseProjectConfig,
    ) -> List[Branch]:
        release_id = get_release_id(context)
        repo = context.get_repo_from_url(dep.github)
        if not repo:
            raise DependencyResolutionError(
                f"Unable to access GitHub repository for {dep.github}"
            )

        try:
            remote_branch_prefix = find_repo_feature_prefix(repo)
        except Exception:
            context.logger.info(
                f"Could not find feature branch prefix or commit-status context for {repo.clone_url}. Unable to resolve packages."
            )
            return []

        # We will check at least the release branch corresponding to our release id.
        # We may be configured to check backwards on release branches.
        release_branches = []
        for i in range(self.branch_offset_start, self.branch_offset_end):
            remote_matching_branch = construct_release_branch_name(
                remote_branch_prefix, str(release_id - i)
            )
            try:
                release_branches.append(repo.branch(remote_matching_branch))
            except NotFoundError:
                context.logger.info(f"Remote branch {remote_matching_branch} not found")
                pass

        return release_branches


class GitHubReleaseBranchCommitStatusResolver(AbstractGitHubReleaseBranchResolver):
    """Resolver that identifies a ref by finding a beta 2GP package version
    in a commit status on a `feature/NNN` release branch."""

    name = "GitHub Release Branch Commit Status Resolver"
    commit_status_context = "2gp_context"
    commit_status_default = "Build Feature Test Package"
    branch_offset_start = 0
    branch_offset_end = 1


class GitHubReleaseBranchUnlockedResolver(AbstractGitHubReleaseBranchResolver):
    """Resolver that identifies a ref by finding an unlocked package version
    in a commit status on a `feature/NNN` release branch."""

    name = "GitHub Release Branch Unlocked Commit Status Resolver"
    commit_status_context = "unlocked_context"
    commit_status_default = "Build Unlocked Test Package"
    branch_offset_start = 0
    branch_offset_end = 1


class GitHubPreviousReleaseBranchCommitStatusResolver(
    AbstractGitHubReleaseBranchResolver
):
    """Resolver that identifies a ref by finding a beta 2GP package version
    in a commit status on a `feature/NNN` release branch that is earlier
    than the matching local release branch."""

    name = "GitHub Previous Release Branch Commit Status Resolver"
    commit_status_context = "2gp_context"
    commit_status_default = "Build Feature Test Package"
    branch_offset_start = 1
    branch_offset_end = 3


class GitHubPreviousReleaseBranchUnlockedResolver(AbstractGitHubReleaseBranchResolver):
    """Resolver that identifies a ref by finding an unlocked package version
    in a commit status on a `feature/NNN` release branch that is earlier
    than the matching local release branch."""

    name = "GitHub Previous Release Branch Unlocked Commit Status Resolver"
    commit_status_context = "unlocked_context"
    commit_status_default = "Build Unlocked Test Package"
    branch_offset_start = 1
    branch_offset_end = 3


class AbstractGitHubExactMatchCommitStatusResolver(
    AbstractGitHubCommitStatusPackageResolver, abc.ABC
):
    """Abstract base class for resolvers that identify a ref by finding a package version
    in a commit status on a branch whose name matches the local branch."""

    def get_branches(
        self,
        dep: BaseGitHubDependency,
        context: BaseProjectConfig,
    ) -> List[Branch]:
        repo = context.get_repo_from_url(dep.github)
        if not repo:
            raise DependencyResolutionError(
                f"Unable to access GitHub repository for {dep.github}"
            )

        try:
            remote_branch_prefix = find_repo_feature_prefix(repo)
        except Exception:
            context.logger.info(
                f"Could not find feature branch prefix or commit-status context for {repo.clone_url}. Unable to resolve package."
            )
            return []

        # Attempt exact match
        try:
            branch = get_feature_branch_name(
                context.repo_branch, context.project__git__prefix_feature
            )
            release_branch = repo.branch(f"{remote_branch_prefix}{branch}")
        except Exception:
            context.logger.info(f"Exact-match branch not found for {repo.clone_url}.")
            return []

        return [release_branch]


class GitHubExactMatch2GPResolver(AbstractGitHubExactMatchCommitStatusResolver):
    """Resolver that identifies a ref by finding a 2GP package version
    in a commit status on a branch whose name matches the local branch."""

    name = "GitHub Exact-Match Commit Status Resolver"
    commit_status_context = "2gp_context"
    commit_status_default = "Build Feature Test Package"


class GitHubExactMatchUnlockedCommitStatusResolver(
    AbstractGitHubExactMatchCommitStatusResolver
):
    """Resolver that identifies a ref by finding an unlocked package version
    in a commit status on a branch whose name matches the local branch."""

    name = "GitHub Exact-Match Unlocked Commit Status Resolver"
    commit_status_context = "unlocked_context"
    commit_status_default = "Build Unlocked Test Package"


class AbstractGitHubDefaultBranchCommitStatusResolver(
    AbstractGitHubCommitStatusPackageResolver, abc.ABC
):
    """Abstract base class for resolvers that identify a ref by finding a beta package version
    in a commit status on the repo's default branch."""

    def get_branches(
        self,
        dep: BaseGitHubDependency,
        context: BaseProjectConfig,
    ) -> List[Branch]:
        repo = context.get_repo_from_url(dep.github)

        return [repo.branch(repo.default_branch)]


class GitHubDefaultBranch2GPResolver(AbstractGitHubDefaultBranchCommitStatusResolver):
    name = "GitHub Default Branch Commit Status Resolver"
    commit_status_context = "2gp_context"
    commit_status_default = "Build Feature Test Package"


class GitHubDefaultBranchUnlockedCommitStatusResolver(
    AbstractGitHubDefaultBranchCommitStatusResolver
):
    name = "GitHub Default Branch Unlocked Commit Status Resolver"
    commit_status_context = "unlocked_context"
    commit_status_default = "Build Unlocked Test Package"


RESOLVER_CLASSES = {
    DependencyResolutionStrategy.STATIC_TAG_REFERENCE: GitHubTagResolver,
    DependencyResolutionStrategy.COMMIT_STATUS_EXACT_BRANCH: GitHubExactMatch2GPResolver,
    DependencyResolutionStrategy.COMMIT_STATUS_RELEASE_BRANCH: GitHubReleaseBranchCommitStatusResolver,
    DependencyResolutionStrategy.COMMIT_STATUS_PREVIOUS_RELEASE_BRANCH: GitHubPreviousReleaseBranchCommitStatusResolver,
    DependencyResolutionStrategy.COMMIT_STATUS_DEFAULT_BRANCH: GitHubDefaultBranch2GPResolver,
    DependencyResolutionStrategy.BETA_RELEASE_TAG: GitHubBetaReleaseTagResolver,
    DependencyResolutionStrategy.RELEASE_TAG: GitHubReleaseTagResolver,
    DependencyResolutionStrategy.UNMANAGED_HEAD: GitHubUnmanagedHeadResolver,
    DependencyResolutionStrategy.UNLOCKED_EXACT_BRANCH: GitHubExactMatchUnlockedCommitStatusResolver,
    DependencyResolutionStrategy.UNLOCKED_RELEASE_BRANCH: GitHubReleaseBranchUnlockedResolver,
    DependencyResolutionStrategy.UNLOCKED_PREVIOUS_RELEASE_BRANCH: GitHubPreviousReleaseBranchUnlockedResolver,
    DependencyResolutionStrategy.UNLOCKED_DEFAULT_BRANCH: GitHubDefaultBranchUnlockedCommitStatusResolver,
}


## External API


def get_resolver(
    strategy: DependencyResolutionStrategy, dependency: DynamicDependency
) -> Optional[AbstractResolver]:
    """Return an instance of a resolver class capable of applying the specified
    resolution strategy to the dependency."""
    # This will be fleshed out when further types of DynamicDependency are added.

    return RESOLVER_CLASSES[strategy]()


def get_resolver_stack(
    context: BaseProjectConfig, name: str
) -> List[DependencyResolutionStrategy]:
    """Return a sequence of resolution strategies identified by the given `name`,
    which can be either a named strategy from project__dependency_resolutions__resolution_strategies
    or an alias like `production`."""
    resolutions = context.project__dependency_resolutions
    stacks = context.project__dependency_resolutions__resolution_strategies

    if name in resolutions and name != "resolution_strategies":
        name = resolutions[name]

    if stacks and name in stacks:
        return [DependencyResolutionStrategy(n) for n in stacks[name]]

    raise CumulusCIException(f"Resolver stack {name} was not found.")


def dependency_filter_ignore_deps(ignore_deps: List[dict]) -> Callable:
    ignore_github = [d["github"] for d in ignore_deps if "github" in d]
    ignore_namespace = [d["namespace"] for d in ignore_deps if "namespace" in d]

    def should_include(some_dep: Dependency) -> bool:

        if isinstance(some_dep, PackageNamespaceVersionDependency):
            return some_dep.namespace not in ignore_namespace
        if isinstance(some_dep, BaseGitHubDependency):
            return some_dep.github not in ignore_github

        return True

    return should_include


def get_static_dependencies(
    context: BaseProjectConfig,
    dependencies: Optional[List[Dependency]] = None,
    resolution_strategy: Optional[str] = None,
    strategies: Optional[List[DependencyResolutionStrategy]] = None,
    filter_function: Optional[Callable] = None,
    pins: Optional[List[DependencyPin]] = None,
) -> List[StaticDependency]:
    """Resolves the dependencies of a CumulusCI project
    to convert dynamic GitHub dependencies into static dependencies
    by inspecting the referenced repositories.

    Keyword arguments:
    :param context: a project config
    :param dependencies: an optional list of dependencies to resolve
                         (defaults to project__dependencies from the project_config)
    :param resolution_strategy: name of a resolution strategy or stack
    :param strategies: list of resolution strategies to use
                       (specify this or resolution_strategy but not both)
    :param filter_function: if provided, call the function with each dependency
                            (including transitive ones) encountered, and include
                            those for which True is returned.
    """
    if dependencies is None:
        dependencies = parse_dependencies(context.project__dependencies)
    if pins is None:
        pins = parse_pins(context.lookup("project__dependency_pins"))
    assert (resolution_strategy is None and strategies is not None) or (
        strategies is None and resolution_strategy is not None
    ), "Expected resolution_strategy or strategies but not both"
    if resolution_strategy:
        strategies = get_resolver_stack(context, resolution_strategy)
    if filter_function is None:
        filter_function = lambda x: True  # noqa: E731

    while any(not d.is_flattened or not d.is_resolved for d in dependencies):
        for d in dependencies:
            if isinstance(d, DynamicDependency) and not d.is_resolved:
                # Finish resolving the dependency using our given strategies.
                d.resolve(context, strategies, pins)

        def unique(it: Iterable):
            seen = set()

            for each in it:
                if each not in seen:
                    seen.add(each)
                    yield each

        dependencies = list(
            unique(
                itertools.chain(
                    *list(
                        d.flatten(context) for d in dependencies if filter_function(d)
                    )
                ),
            )
        )

    # Make sure, if we had no flattening or resolving to do, that we apply the ignore list.
    # Type is guaranteed via the logic above.
    return [d for d in dependencies if filter_function(d)]  # type: ignore


def resolve_dependency(
    dependency: DynamicDependency,
    context: BaseProjectConfig,
    strategies: List[DependencyResolutionStrategy],
    pins: Optional[List[DependencyPin]] = None,
):
    """Resolve a DynamicDependency that is not pinned to a specific version into one that is.

    If successful, sets `dependency.ref` and optionally `dependency.package_dependency`
    (if a package release is found).

    Otherwise raises DependencyResolutionError.
    """

    if dependency.is_resolved:
        return

    for s in strategies:
        resolver = get_resolver(s, dependency)

        if resolver and resolver.can_resolve(dependency, context):
            try:
                dependency.ref, dependency.package_dependency = resolver.resolve(
                    dependency, context
                )
                if dependency.package_dependency:
                    try:
                        dependency.package_dependency.password_env_name = (
                            dependency.password_env_name
                        )
                    except AttributeError:  # pragma: no cover
                        pass
                if dependency.ref:
                    break
            except DependencyResolutionError:
                context.logger.info(
                    f"Resolution strategy {s} failed for dependency {dependency}."
                )

    if not dependency.ref:
        raise DependencyResolutionError(f"Unable to resolve dependency {dependency}")



import faker
from robot.libraries.BuiltIn import RobotNotRunningError


class FakerMixin:
    """Mixin class which provides support for the faker library"""

    def __init__(self):
        super().__init__()
        self._faker = faker.Faker("en_US")
        try:
            self.builtin.set_global_variable("${faker}", self._faker)
        except RobotNotRunningError:
            # this only happens during unit tests, and we don't care.
            pass

    def set_faker_locale(self, locale):
        """Set the locale for fake data

        This sets the locale for all calls to the ``Faker`` keyword
        and ``${faker}`` variable. The default is en_US

        For a list of supported locales see
        [https://faker.readthedocs.io/en/master/locales.html|Localized Providers]
        in the Faker documentation.

        Example

        | Set Faker Locale    fr_FR
        | ${french_address}=  Faker  address

        """
        try:
            self._faker = faker.Faker(locale)
        except AttributeError:
            raise Exception(f"Unknown locale for fake data: '{locale}'")

    def get_fake_data(self, fake, *args, **kwargs):
        """Return fake data

        This uses the [https://faker.readthedocs.io/en/master/|Faker]
        library to provide fake data in a variety of formats (names,
        addresses, credit card numbers, dates, phone numbers, etc) and
        locales (en_US, fr_FR, etc).

        The _fake_ argument is the name of a faker property such as
        ``first_name``, ``address``, ``lorem``, etc. Additional
        arguments depend on type of data requested. For a
        comprehensive list of the types of fake data that can be
        generated see
        [https://faker.readthedocs.io/en/master/providers.html|Faker
        providers] in the Faker documentation.

        The return value is typically a string, though in some cases
        some other type of object will be returned. For example, the
        ``date_between`` fake returns a
        [https://docs.python.org/3/library/datetime.html#date-objects|datetime.date
        object]. Each time a piece of fake data is requested it will
        be regenerated, so that multiple calls will usually return
        different data.

        This keyword can also be called using robot's extended variable
        syntax using the variable ``${faker}``. In such a case, the
        data being asked for is a method call and arguments must be
        enclosed in parentheses and be quoted. Arguments should not be
        quoted when using the keyword.

        To generate fake data for a locale other than en_US, use
        the keyword ``Set Faker Locale`` prior to calling this keyword.

        Examples

        | # Generate a fake first name
        | ${first_name}=  Get fake data  first_name

        | # Generate a fake date in the default format
        | ${date}=  Get fake data  date

        | # Generate a fake date with an explicit format
        | ${date}=  Get fake data  date  pattern=%Y-%m-%d

        | # Generate a fake date using extended variable syntax
        | Input text  //input  ${faker.date(pattern='%Y-%m-%d')}

        """
        try:
            return self._faker.format(fake, *args, **kwargs)
        except AttributeError:
            raise Exception(f"Unknown fake data request: '{fake}'")



from pathlib import Path

from cumulusci.core.config.org_config import OrgConfig
from cumulusci.core.datasets import Dataset
from cumulusci.core.exceptions import TaskOptionsError
from cumulusci.salesforce_api.filterable_objects import OPT_IN_ONLY
from cumulusci.salesforce_api.org_schema import Filters, get_org_schema
from cumulusci.tasks.salesforce.BaseSalesforceApiTask import BaseSalesforceApiTask


class CaptureSampleData(BaseSalesforceApiTask):
    """Capture contextual sample data based on repo."""

    task_options = {
        "dataset": {
            "description": (
                "The name of the dataset. If none is provided, it will use 'default'. "
                "Names that match scratch org config names (such as 'dev', 'qa') will be loaded "
                "into those orgs by the CumulusCI default org setup flows."
            )
        },
        "extraction_definition": {
            "description": "A file describing what to be extracted. "
            "Defaults to `datasets/{datasetname}/{datasetname}.extract.yml` if it exists."
        },
        "loading_rules": {
            "description": (
                "Path to .load.yml file containing rules to use when loading the mapping. "
                "Defaults to`datasets/{datasetname}/{datasetname}.load.yml` if it exists. "
                "Multiple files can be comma separated."
            )
        },
    }

    org_config: OrgConfig

    def _init_options(self, kwargs):
        super()._init_options(kwargs)
        self.options.setdefault("dataset", "default")

    def _run_task(self):
        name = self.options["dataset"]
        if extraction_definition := self.options.get("extraction_definition"):
            extraction_definition = Path(extraction_definition)
            if not extraction_definition.exists():
                raise TaskOptionsError(f"Cannot find {extraction_definition}")

        if loading_rules := self.options.get("loading_rules"):
            loading_rules = Path(loading_rules)
            if not loading_rules.exists():
                raise TaskOptionsError(f"Cannot find {loading_rules}")

        with get_org_schema(
            self.sf,
            self.org_config,
            include_counts=True,
            filters=[Filters.extractable, Filters.createable],
        ) as schema, Dataset(
            name,
            self.project_config,
            self.sf,
            self.org_config,
            schema,
        ) as dataset:
            if not dataset.path.exists():
                dataset.create()
                verb = "Created"
            else:
                verb = "Updated"
            opt_in_only = [f["name"] for f in self.tooling.describe()["sobjects"]]  # type: ignore
            opt_in_only += OPT_IN_ONLY

            self.return_values = dataset.extract(
                {}, self.logger, extraction_definition, opt_in_only, loading_rules
            )
            self.logger.info(f"{verb} dataset '{name}' in 'datasets/{name}'")
            return self.return_values



import typing as T
from contextlib import contextmanager
from dataclasses import dataclass
from io import StringIO
from logging import Logger, getLogger
from pathlib import Path
from shutil import rmtree
from tempfile import TemporaryDirectory

import yaml
from simple_salesforce import Salesforce
from snowfakery.cci_mapping_files.declaration_parser import (
    SObjectRuleDeclaration,
    SObjectRuleDeclarationFile,
)

from cumulusci.core import exceptions as exc
from cumulusci.core.config import BaseProjectConfig, OrgConfig, TaskConfig
from cumulusci.salesforce_api.org_schema import Filters, Schema, get_org_schema
from cumulusci.tasks.bulkdata.extract import ExtractData
from cumulusci.tasks.bulkdata.extract_dataset_utils.extract_yml import (
    ExtractDeclaration,
    ExtractRulesFile,
)
from cumulusci.tasks.bulkdata.extract_dataset_utils.synthesize_extract_declarations import (
    SimplifiedExtractDeclaration,
    flatten_declarations,
)
from cumulusci.tasks.bulkdata.generate_mapping_utils.extract_mapping_file_generator import (
    create_extract_mapping_file_from_declarations,
)
from cumulusci.tasks.bulkdata.generate_mapping_utils.generate_mapping_from_declarations import (
    create_load_mapping_file_from_extract_declarations,
)
from cumulusci.tasks.bulkdata.load import LoadData
from cumulusci.tasks.bulkdata.snowfakery import Snowfakery

DEFAULT_LOGGER = getLogger(__file__)


@dataclass
class Dataset:
    name: str
    project_config: BaseProjectConfig
    sf: Salesforce
    org_config: OrgConfig
    schema: T.Optional[Schema] = None
    initialized = False

    @property
    def path(self) -> Path:
        return Path(self.project_config.repo_root or "") / "datasets" / self.name

    @property
    def extract_file(self) -> Path:
        return self.path / f"{self.name}.extract.yml"

    @property
    def load_rules_file(self) -> Path:
        # to be merged into the mapping file when it is created on extract
        return self.path / f"{self.name}.load.yml"

    @property
    def mapping_file(self) -> Path:
        return self.path / f"{self.name}.mapping.yml"

    @property
    def data_file(self) -> Path:
        return self.path / f"{self.name}.dataset.sql"

    @property
    def snowfakery_recipe(self) -> Path:
        return self.path / f"{self.name}.recipe.yml"

    def exists(self) -> bool:
        return self.path.exists()

    def delete(self):
        rmtree(str(self.path))

    def __enter__(self, *args, **kwargs):
        if self.schema is None:
            self.schema_context = self._get_org_schema()
            self.schema = self.schema_context.__enter__(*args, **kwargs)  # type: ignore
        else:
            self.schema_context = None
        self.initialized = True
        return self

    def __exit__(self, *args, **kwargs):
        if self.schema_context:
            self.schema_context.__exit__(*args, **kwargs)  # type: ignore

    def create(self):
        assert (
            self.initialized
        ), "You must open this context manager. e.g. `with Dataset() as dataset`"

        if not self.path.exists():
            self.path.mkdir()

        self.extract_file.write_text(DEFAULT_EXTRACT_DATA)
        decls = ExtractRulesFile.parse_extract(StringIO(DEFAULT_EXTRACT_DATA))

        decls = list(decls.values())

    def _get_org_schema(self):
        return get_org_schema(
            self.sf,
            self.org_config,
            include_counts=True,
            filters=[Filters.extractable, Filters.createable, Filters.populated],
        )

    def _save_load_mapping(
        self,
        decls: T.Sequence[ExtractDeclaration],
        opt_in_only: T.Sequence[str] = (),
        loading_rules: T.Sequence[SObjectRuleDeclaration] = (),
    ) -> None:
        assert isinstance(self.schema, Schema)
        mapping_data = create_load_mapping_file_from_extract_declarations(
            decls, self.schema, opt_in_only, loading_rules
        )
        with self.mapping_file.open("w") as f:
            f.write(EDIT_MAPPING_WARNING)
            yaml.safe_dump(mapping_data, f, sort_keys=False)

    @contextmanager
    def temp_extract_mapping(
        self,
        schema,
        extraction_definition: T.Optional[Path],
        opt_in_only: T.Sequence[str],
    ):
        with TemporaryDirectory() as t:
            t = Path(t)
            if extraction_definition:
                assert extraction_definition.exists(), "Cannot find extract mapping {f}"
                self.extract_file.write_text(extraction_definition.read_text())
            else:
                extraction_definition = self.extract_file

                if not self.extract_file.exists():  # pragma: no cover
                    self.extract_file.write_text(DEFAULT_EXTRACT_DATA)
            decls = ExtractRulesFile.parse_extract(extraction_definition)

            extract_mapping = t / "extract.mapping.yml"
            with extract_mapping.open("w") as f:
                yaml.safe_dump(
                    create_extract_mapping_file_from_declarations(
                        list(decls.values()), schema, opt_in_only
                    ),
                    f,
                )
            yield extract_mapping, decls

    def extract(
        self,
        options: T.Optional[T.Dict] = None,
        logger: T.Optional[Logger] = None,
        extraction_definition: T.Optional[Path] = None,
        opt_in_only: T.Sequence[str] = (),
        loading_rules_file: T.Optional[Path] = None,
    ):
        options = options or {}
        logger = logger or DEFAULT_LOGGER
        with self.temp_extract_mapping(
            self.schema, extraction_definition, opt_in_only
        ) as (
            extract_mapping,
            decls,
        ):
            task = _make_task(
                ExtractData,
                project_config=self.project_config,
                org_config=self.org_config,
                sql_path=self.data_file,
                mapping=str(extract_mapping),
            )
            task()
        loading_rules = self._parse_loading_rules_file(loading_rules_file)

        self._save_load_mapping(list(decls.values()), opt_in_only, loading_rules)
        return task.return_values

    def _parse_loading_rules_file(
        self, loading_rules_file: T.Optional[Path]
    ) -> T.List[SObjectRuleDeclaration]:
        """Parse a loading rules file if provided, else try to parse
        <datasets/dataset/dataset.load.yml>"""
        if loading_rules_file:
            assert loading_rules_file.exists()
        else:
            if self.load_rules_file.exists():
                loading_rules_file = self.load_rules_file

        if loading_rules_file:
            with loading_rules_file.open() as f:
                parse_result = SObjectRuleDeclarationFile.parse_from_yaml(f)
                loading_rules = parse_result.sobject_declarations
        else:
            loading_rules = []

        return loading_rules

    def _snowfakery_dataload(self, options: T.Dict, logger: Logger) -> T.Dict:
        subtask_config = TaskConfig(
            {"options": {**options, "recipe": str(self.snowfakery_recipe)}}
        )
        subtask = Snowfakery(
            project_config=self.project_config,
            task_config=subtask_config,
            org_config=self.org_config,
            logger=logger,
        )
        subtask()
        return subtask.return_values

    def load(
        self, options: T.Optional[T.Dict] = None, logger: T.Optional[Logger] = None
    ):
        options = options or {}
        logger = logger or DEFAULT_LOGGER
        if self.data_file.exists():
            self._sql_dataload(options)
        elif self.snowfakery_recipe.exists():
            self._snowfakery_dataload(options, logger)
        else:  # pragma: no cover
            raise exc.BulkDataException(
                f"Dataset has no SQL ({self.data_file}) "
                "or recipe ({self.snowfakery_recipe})"
            )

    def _sql_dataload(self, options: T.Dict):

        task = _make_task(
            LoadData,
            project_config=self.project_config,
            org_config=self.org_config,
            sql_path=self.data_file,
            mapping=str(self.mapping_file),
            **options,
        )
        task()

    def read_schema_subset(self) -> T.Dict[str, T.List[str]]:
        assert isinstance(self.schema, Schema)
        decls = ExtractRulesFile.parse_extract(self.extract_file)
        flattened = flatten_declarations(list(decls.values()), self.schema)
        return {obj.sf_object: obj.fields for obj in flattened}  # type: ignore

    def read_which_fields_selected(self) -> T.Dict[str, T.Dict[str, bool]]:
        assert isinstance(self.schema, Schema)
        selected = self.read_schema_subset()
        selected_fields = {
            (obj, field) for obj, fields in selected.items() for field in fields
        }
        out = {}
        for objname, obj in self.schema.items():
            objrepr = out.setdefault(objname, {})
            for field in obj.fields.keys():
                is_selected = (objname, field) in selected_fields
                objrepr[field] = is_selected
        return out

    def update_schema_subset(
        self,
        objs: T.Dict[str, T.List[str]],
        opt_in_only: T.Sequence[str] = (),
    ):
        # TODO: Test round-tripping through this carefully...especially
        #       for weird objects like WorkBadgeDefinitions
        objs_dict = {name: {"fields": fields} for name, fields in objs.items()}
        with self.extract_file.open("w") as f:
            data = {"extract": objs_dict}
            yaml.safe_dump(data, f, sort_keys=False)
        decls = [
            SimplifiedExtractDeclaration(sf_object=name, fields=fields)
            for name, fields in objs.items()
        ]
        self._save_load_mapping(decls, opt_in_only)


def _make_task(task_class, project_config, org_config, **options):
    task_config = TaskConfig({"options": options})
    return task_class(project_config, task_config, org_config)


DEFAULT_EXTRACT_DATA = """
extract:
    OBJECTS(ALL):
        fields: FIELDS(ALL)
"""

EDIT_MAPPING_WARNING = """# Editing this file is usually not recommended because it will
# be overwritten the next time you re-capture this data.
#
# You can change this file's contents permanently by creating a
# .load.yml file and re-capturing:
#
#  https://cumulusci.readthedocs.io/en/stable/data.html#extracting-and-loading-sample-datasets
"""



import json

import click
from rich.console import Console

from cumulusci.cli.ui import CliTable
from cumulusci.core.metadeploy.plans import get_frozen_steps
from cumulusci.utils.yaml.cumulusci_yml import Plan

from .runtime import pass_runtime


@click.group("plan", help="Commands for getting information about MetaDeploy plans")
def plan():
    pass  # pragma: no cover


@plan.command(name="list")
@click.option(
    "--json", "print_json", is_flag=True, help="Return the list of plans in JSON format"
)
@pass_runtime(require_project=True, require_keychain=True)
def plan_list(runtime, print_json):
    """List available plans for the current context.

    If --json is specified, the data will be a list of plans where
    each plan is a dictionary with the following keys:
    name, group, title, slug, tier

    """

    plans = runtime.project_config.plans or {}

    def _sort_func(plan):
        """Used to sort first by tier (primary, secondary, additional)
        and then by name"""
        name, config = plan
        tier = config.get("tier", "primary")
        tiers = Plan.schema()["properties"]["tier"]["enum"]
        tier_index = tiers.index(tier) if tier in tiers else 99
        return f"{tier_index} {name}"

    columns = ["name", "title", "slug", "tier"]
    raw_data = [
        dict(name=plan_name, **{key: plan_config.get(key, "") for key in columns[1:]})
        for plan_name, plan_config in sorted(plans.items(), key=_sort_func)
    ]

    if print_json:
        click.echo(json.dumps(raw_data))
        return

    data = [[name.title() for name in columns]]
    data.extend([list(row.values()) for row in raw_data])
    console = Console()
    table = CliTable(data=data)
    console.print(table)

    click.echo(
        "Use "
        + click.style("cci plan info <plan_name>", bold=True)
        + " to get more information about a task."
    )


@plan.command(name="info")
@click.argument("plan_name")
@click.option(
    "--messages", "messages_only", is_flag=True, help="Show only plan messages"
)
@pass_runtime(require_project=True, require_keychain=True)
def plan_info(runtime, plan_name, messages_only):
    """
    Displays information for a MetaDeploy plan.
    """

    plans = runtime.project_config.plans or {}

    if plan_name not in plans:
        raise click.UsageError(
            f"Unknown plan '{plan_name}'. To view available plans run: `cci plan list`"
        )

    plan_config = plans[plan_name]

    console = Console()

    raw_data = {
        "title": plan_config["title"],
        "yaml_key": plan_name,
        "slug": plan_config["slug"],
        "tier": plan_config["tier"],
        "hidden": not plan_config.get("is_listed", "False"),
        "preflight_message": plan_config.get("preflight_message", ""),
        "post_install_message": plan_config.get("post_install_message", ""),
        "error_message": plan_config.get("error_message", ""),
        "checks": [
            [check.get("action", ""), check.get("message", ""), check.get("when", "")]
            for check in plan_config.get("checks", [])
        ],
        "steps": {},  # expensive to compute so we'll do later
    }

    messages_table = CliTable(
        title="Messages",
        data=[
            ["Type", "Message"],
            ["Title", raw_data["title"]],
            ["Preflight", raw_data["preflight_message"]],
            ["Post-install", raw_data["post_install_message"]],
            ["Error", raw_data["error_message"]],
        ],
    )

    if messages_only:
        console.print(messages_table)
        return

    config_table = CliTable(
        title="Config",
        data=[
            ["Key", "Value"],
            ["YAML Key", raw_data["yaml_key"]],
            ["Slug", raw_data["slug"]],
            ["Tier", raw_data["tier"]],
            ["Hidden?", raw_data["hidden"]],
        ],
    )

    plan_preflight_checks_table = CliTable(
        title="Plan Preflights",
        data=[
            ["Action", "Message", "When"],
            *raw_data["checks"],
        ],
    )

    steps = get_frozen_steps(runtime.project_config, plan_config)
    raw_data["steps"] = [
        [
            step_n + 1,
            step["name"],
            step["is_required"],
            step.get("is_recommended", True),
        ]
        for step_n, step in enumerate(steps)
    ]
    raw_data["steps_preflight_checks"] = [
        [
            step_n + 1,
            check.get("action", ""),
            check.get("message", ""),
            check.get("when", ""),
        ]
        for step_n, step in enumerate(steps)
        for check in step["task_config"]["checks"]
    ]

    step_preflight_checks_table = CliTable(
        title="Step Preflights",
        data=[
            ["Step", "Action", "Message", "When"],
            *raw_data["steps_preflight_checks"],
        ],
    )

    CliTable.PICTOGRAM_TRUE = "Yes"
    CliTable.PICTOGRAM_FALSE = "No"
    steps_table = CliTable(
        title="Steps",
        data=[
            ["Step", "Name", "Required", "Recommended"],
            *raw_data["steps"],
        ],
    )

    console.print(config_table)
    console.print(messages_table)
    console.print(plan_preflight_checks_table)
    console.print(step_preflight_checks_table)
    console.print(steps_table)



import tempfile
from datetime import datetime
from pathlib import Path

import github3.exceptions

from cumulusci.core.exceptions import (
    CumulusCIException,
    GithubException,
    TaskOptionsError,
)
from cumulusci.core.github import get_tag_by_name
from cumulusci.core.utils import process_bool_arg, process_list_arg
from cumulusci.tasks.github.base import BaseGithubTask
from cumulusci.tasks.github.util import CommitDir
from cumulusci.utils import download_extract_github_from_repo
from cumulusci.utils.git import split_repo_url


class PublishSubtree(BaseGithubTask):
    task_options = {  # TODO: should use `class Options instead`
        "repo_url": {"description": "The url to the public repo", "required": True},
        "branch": {
            "description": "The branch to update in the target repo",
            "required": True,
        },
        "version": {
            "description": "(Deprecated >= 3.42.0) Only the values of 'latest' and 'latest_beta' are acceptable. "
            "Required if 'ref' or 'tag_name' is not set. This will override tag_name if it is provided."
        },
        "tag_name": {
            "description": "The name of the tag that should be associated with this release. "
            "Values of 'latest' and 'latest_beta' are also allowed. "
            "Required if 'ref' or 'version' is not set."
        },
        "ref": {
            "description": "The git reference to publish.  Takes precedence over 'version' and 'tag_name'. "
            "Required if 'tag_name' is not set."
        },
        "include": {
            "description": "A list of paths from repo root to include. Directories must end with a trailing slash."
        },
        "renames": {
            "description": "A list of paths to rename in the target repo, given as `local:` `target:` pairs."
        },
        "create_release": {
            "description": "If True, create a release in the public repo.  Defaults to False"
        },
        "release_body": {
            "description": "If True, the entire release body will be published to the public repo.  Defaults to False"
        },
        "dry_run": {
            "description": "If True, skip creating Github data.  Defaults to False"
        },
    }

    def _init_options(self, kwargs):
        super()._init_options(kwargs)
        self.options["include"] = process_list_arg(
            self.options.get(
                "include", ["datasets/", "documentation/", "tasks/", "unpackaged/"]
            )
        )

        self.options["renames"] = self._process_renames(self.options.get("renames", []))

        if "version" in self.options:  # pragma: no cover
            self.logger.warning(
                "The `version` option is deprecated. Please use the `tag_name` option instead."
            )
            if self.options["version"] not in ("latest", "latest_beta"):
                raise TaskOptionsError(
                    f"Only `latest` and `latest_beta` are valid values for the `version` option. Found: {self.options['version']}"
                )

        if (
            "ref" not in self.options
            and "tag_name" not in self.options
            and "version" not in self.options
        ):
            raise TaskOptionsError("Either `ref` or `tag_name` option is required.")

        self.options["create_release"] = process_bool_arg(
            self.options.get("create_release", False)
        )
        self.options["release_body"] = process_bool_arg(
            self.options.get("release_body", False)
        )
        self.options["dry_run"] = process_bool_arg(self.options.get("dry_run", False))

    def _process_renames(self, renamed_paths):
        """
        For each entry in renames, any renames and store them
        in self.local_to_target_paths.
        """
        if not renamed_paths:
            return {}

        is_list_of_dicts = all(isinstance(pair, dict) for pair in renamed_paths)
        dicts_have_correct_keys = is_list_of_dicts and all(
            {"local", "target"} == pair.keys() for pair in renamed_paths
        )

        ERROR_MSG = (
            "Renamed paths must be a list of dicts with `local:` and `target:` keys."
        )
        if not dicts_have_correct_keys:
            raise TaskOptionsError(ERROR_MSG)

        local_to_target_paths = {}

        for rename in renamed_paths:
            local_path = rename.get("local")
            target_path = rename.get("target")

            if local_path and target_path:
                local_to_target_paths[local_path] = target_path
            else:
                raise TaskOptionsError(ERROR_MSG)

        return local_to_target_paths

    def _get_target_repo_api(self):
        owner, name = split_repo_url(self.options["repo_url"])
        gh = self.project_config.get_github_api(self.options["repo_url"])
        return gh.repository(owner, name)

    def _run_task(self):
        self.target_repo = self._get_target_repo_api()
        self._set_ref()

        with tempfile.TemporaryDirectory() as target:
            self._download_repo_and_extract(target)
            self._rename_files(target)
            commit = self._create_commit(target)
            if commit and self.options["create_release"]:
                self._create_release(target, commit.sha)

    def _set_ref(self):
        if "ref" in self.options:
            self.ref = self.options["ref"]

        elif "version" in self.options:
            get_beta = self.options.get("version") == "latest_beta"
            self.tag_name = self.project_config.get_latest_tag(beta=get_beta)
            self.ref = f"tags/{self.tag_name}"

        elif "tag_name" in self.options:
            if self.options["tag_name"] in ("latest", "latest_beta"):
                get_beta = self.options["tag_name"] == "latest_beta"
                tag_name = self.project_config.get_latest_tag(beta=get_beta)
            else:
                tag_name = self.options["tag_name"]

            self.tag_name = tag_name
            self.ref = f"tags/{self.tag_name}"

        else:  # pragma: no cover
            raise CumulusCIException("No ref, version, or tag_name present")

    def _download_repo_and_extract(self, path):
        zf = download_extract_github_from_repo(self.get_repo(), ref=self.ref)
        included_members = self._filter_namelist(
            includes=self.options["include"], namelist=zf.namelist()
        )
        zf.extractall(path=path, members=included_members)

    def _filter_namelist(self, includes, namelist):
        """
        Filter a zipfile namelist, handling any included directory filenames missing
        a trailing slash.
        """
        included_dirs = []
        zip_dirs = [
            filename.rstrip("/") for filename in namelist if filename.endswith("/")
        ]

        for name in includes:
            if name.endswith("/"):
                included_dirs.append(name)
            elif name in zip_dirs:
                # append a trailing slash to avoid partial matches
                included_dirs.append(name + "/")

        return list(
            {
                name
                for name in namelist
                if name.startswith(tuple(included_dirs)) or name in includes
            }
        )

    def _rename_files(self, zip_dir):
        for local_name, target_name in self.options["renames"].items():
            local_path = Path(zip_dir, local_name)
            if local_path.exists():
                target_path = Path(zip_dir, target_name)
                target_path.parent.mkdir(parents=True, exist_ok=True)
                local_path.replace(target_path)

    def _create_commit(self, path):
        committer = CommitDir(self.target_repo, logger=self.logger)
        message = f"Published content from ref {self.ref}"
        if "tag_name" in self.options:
            message += f"\n\nTag {self.tag_name}"
        return committer(
            path,
            self.options["branch"],
            repo_dir="",
            commit_message=message,
            dry_run=self.options["dry_run"],
        )

    def _create_release(self, path, commit):
        repo = self.get_repo()
        tag = get_tag_by_name(repo, self.tag_name)

        # Get the release
        try:
            release = repo.release_from_tag(self.tag_name)
            release_body = release.body if self.options["release_body"] else ""
        except github3.exceptions.NotFoundError:
            message = f"Release for {self.tag_name} not found"
            raise GithubException(message)

        # Check for tag in target repo
        try:
            self.target_repo.ref(f"tags/{self.tag_name}")
        except github3.exceptions.NotFoundError:
            pass
        else:
            message = f"Ref for tag {self.tag_name} already exists in target repo"
            raise GithubException(message)

        # Create the tag
        self.target_repo.create_tag(
            tag=self.tag_name,
            message=tag.message,
            sha=commit,
            obj_type="commit",
            tagger={
                "name": self.github_config.username,
                "email": self.github_config.email,
                "date": f"{datetime.utcnow().isoformat()}Z",
            },
            lightweight=False,
        )

        # Create the release
        self.target_repo.create_release(
            tag_name=self.tag_name,
            name=self.project_config.get_version_for_tag(self.tag_name),
            prerelease=release.prerelease,
            body=release_body,
        )



import typing as T
from itertools import chain

from snowfakery.cci_mapping_files.declaration_parser import SObjectRuleDeclaration

from cumulusci.salesforce_api.org_schema import Schema
from cumulusci.tasks.bulkdata.mapping_parser import MappingStep
from cumulusci.utils.collections import OrderedSet
from cumulusci.utils.iterators import partition

from ..extract_dataset_utils.calculate_dependencies import SObjDependency
from ..extract_dataset_utils.synthesize_extract_declarations import (
    ExtractDeclaration,
    SimplifiedExtractDeclaration,
    flatten_declarations,
)
from .load_mapping_file_generator import generate_load_mapping_file


class SimplifiedExtractDeclarationWithLookups(SimplifiedExtractDeclaration):
    lookups: T.Dict[str, T.Union[str, T.Tuple[str, ...]]]


def create_load_mapping_file_from_extract_declarations(
    decls: T.Sequence[ExtractDeclaration],
    schema: Schema,
    opt_in_only: T.Sequence[str] = (),
    loading_rules: T.Sequence[SObjectRuleDeclaration] = (),
) -> T.Dict[str, dict]:
    """Create a mapping file from Extract declarations"""
    simplified_decls = flatten_declarations(decls, schema, opt_in_only)  # FIXME
    simplified_decls_w_lookups = classify_and_filter_lookups(simplified_decls, schema)
    intertable_dependencies = _discover_dependendencies(simplified_decls_w_lookups)

    def _mapping_step(decl):
        fields = tuple(chain(decl.fields, decl.lookups.keys()))
        return MappingStep(
            sf_object=decl.sf_object,
            fields=dict(zip(fields, fields)),
            # lookups=lookups,      # lookups can be re-created later, for simplicity
        )

    mapping_steps = [_mapping_step(decl) for decl in simplified_decls_w_lookups]

    mappings = generate_load_mapping_file(
        mapping_steps, intertable_dependencies, loading_rules
    )
    return mappings


def _discover_dependendencies(simplified_decls: T.Sequence):
    """Look at all of the lookups in a set of declarations to determine
    what depends on what"""
    intertable_dependencies = OrderedSet()

    for decl in simplified_decls:
        for fieldname, tablenames in decl.lookups.items():
            intertable_dependencies.add(
                SObjDependency(decl.sf_object, tablenames, fieldname)
            )
    return intertable_dependencies


def classify_and_filter_lookups(
    decls: T.Sequence[SimplifiedExtractDeclaration], schema: Schema
) -> T.Sequence[SimplifiedExtractDeclarationWithLookups]:
    """Move lookups into their own field, if they reference a table we're including"""
    referenceable_tables = [decl.sf_object for decl in decls if decl.sf_object]
    # the if statement above is just to shut up the type checker
    return [_add_lookups_to_decl(decl, schema, referenceable_tables) for decl in decls]


def _add_lookups_to_decl(
    decl: SimplifiedExtractDeclaration,
    schema: Schema,
    referenceable_tables: T.Sequence[str],
) -> SimplifiedExtractDeclarationWithLookups:
    """Look at every declaration and check whether any of the fields it refers to
    are actually lookups. If so, synthesize the lookups declarations."""
    sobject_schema_info = schema[decl.sf_object]
    fields, lookups_and_targets = _fields_and_lookups_for_decl(
        decl, sobject_schema_info, referenceable_tables
    )
    new_decl_data = {
        **dict(decl),
        "fields": list(fields),
        "lookups": dict(lookups_and_targets),
    }
    del new_decl_data["fields_"]
    new_decl = SimplifiedExtractDeclarationWithLookups(**new_decl_data)
    return new_decl


def _fields_and_lookups_for_decl(decl, sobject_schema_info, referenceable_tables):
    """Split fields versus lookups for a declaration"""

    def is_lookup(field_name):
        # Record types are not treated as lookup.
        if field_name == "RecordTypeId":
            return False
        schema_info_for_field = sobject_schema_info.fields[field_name]
        target = schema_info_for_field.referenceTo
        return target

    simple_fields, lookups = partition(is_lookup, decl.fields)

    def target_table(field_info):
        target = field_info.referenceTo
        return target

    lookups = list(lookups)

    lookups_and_targets = (
        (lookup, target_table(sobject_schema_info.fields[lookup])) for lookup in lookups
    )
    lookups_and_targets = (
        (lookup, [table for table in tables if table in referenceable_tables])
        for lookup, tables in lookups_and_targets
        if any(table in referenceable_tables for table in tables)
    )
    return simple_fields, lookups_and_targets
